# Copyright 2022 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Polyrhythmic Sequencer
author: Adam Wonak (github.com/awonak)
date: 2022-01-17
labels: polyrhythms, sequencer, triggers

EuroPi version of a Subharmonicon style polyrhythmic sequencer.
Partially inspired by m0wh: https://github.com/m0wh/subharmonicon-sequencer
Demo video: https://youtu.be/vMAVqVQIpW0

Page 1 is the first 4 note sequence, page 2 is the second 4 note sequence, page
3 is the polyrhythms assignable to each sequence. Use knob 1 to select between
the 4 steps and use knob 2 to edit that step. On page 3 there are 4 polyrhythm
options ranging from triggering every 1 beat to every 16 beats. On this page
button 2 assigns which sequence this polyrhythm should apply to. Button 1 will
cycle through the pages. The script needs a clock source in the digital input
to play.

digital_in: clock in
analog_in: unused

knob_1: select step option for current page
knob_2: adjust the value of the selected option

button_1: cycle through editable parameters (seq1, seq1, poly)
button_2: edit current parameter options

output_1: pitch 1
output_2: trigger 1
output_3: trigger logical AND
output_4: pitch 2
output_5: trigger 2
output_6: trigger logical XOR

"""
try:
    # Local development
    from software.firmware.europi import OLED_WIDTH, OLED_HEIGHT, CHAR_HEIGHT
    from software.firmware.europi import din, k1, k2, oled, b1, b2, cv1, cv2, cv3, cv4, cv5, cv6
    from software.firmware.europi_script import EuroPiScript
except ImportError:
    # Device import path
    from europi import *
    from europi_script import EuroPiScript

from collections import namedtuple
import struct
import machine
from utime import ticks_diff, ticks_ms


# Script Constants
MENU_DURATION = 1200

VOLT_PER_OCT = 1 / 12

# fmt: off
NOTES = [
    "C0", "C#0", "D0", "D#0", "E0", "F0", "F#0", "G0", "G#0", "A0", "A#0", "B0",
    "C1", "C#1", "D1", "D#1", "E1", "F1", "F#1", "G1", "G#1", "A1", "A#1", "B1",
    "C2", "C#2", "D2", "D#2", "E2", "F2", "F#2", "G2", "G#2", "A2", "A#2", "B2",
    "C3", "C#3", "D3", "D#3", "E3", "F3", "F#3", "G3", "G#3", "A3", "A#3", "B3",
    "C4", "C#4", "D4", "D#4", "E4", "F4", "F#4", "G4", "G#4", "A4", "A#4", "B4",
]
# fmt: on


class Sequence:
    def __init__(self, notes, pitch_cv, trigger_cv):
        self.notes = notes
        self.pitch_cv = pitch_cv
        self.trigger_cv = trigger_cv
        self.step_index = 0

        # Save state struct
        self.format_string = "4s"
        self.State = namedtuple("State", "note_indexes")

    def set_state(self, state):
        """Update instance variables with given state bytestring."""
        state = self.State(*struct.unpack(self.format_string, state))
        self.notes = [NOTES[n] for n in state.note_indexes]

    def get_state(self):
        """Return state byte string."""
        note_indexes = [NOTES.index(n) for n in self.notes]
        return struct.pack(self.format_string, bytes(note_indexes))

    def _pitch_cv(self, note: str) -> float:
        return NOTES.index(note) * VOLT_PER_OCT

    def _set_pitch(self):
        pitch = self._pitch_cv(self.current_note())
        self.pitch_cv.voltage(pitch)

    def current_note(self) -> str:
        return self.notes[self.step_index]

    def edit_step(self, step: int, note: str):
        """Set the given step to the given note value and update pitch cv out."""
        assert note in NOTES, f"Given note not in available notes: {note}"
        self.notes[step] = note
        self._set_pitch()

    def advance_step(self):
        """Advance the sequence step index."""
        self.step_index = (self.step_index + 1) % len(self.notes)

    def play_next_step(self):
        """Advance the sequence step and play the note."""
        self.advance_step()
        # Set cv output voltage to sequence step pitch.
        self._set_pitch()
        self.trigger_cv.on()

    def reset(self):
        """Reset the sequence back to the first note."""
        self.step_index = 0
        self._set_pitch()
        self.trigger_cv.off()


class PolyrhythmSeq(EuroPiScript):
    pages = ['SEQUENCE 1', 'SEQUENCE 2', 'POLYRHYTHM']
    # Two 4-step melodic sequences.
    seqs = [
        Sequence(["C0", "D#0", "D0", "G0"], cv1, cv2),
        Sequence(["G0", "F0", "D#0", "C0"], cv4, cv5),
    ]
    # 4 editable polyrhythms, assignable to the sequences.
    polys = [8, 3, 2, 5]
    # Indicates which sequences are assigned to each polyrhythm.
    # 0: none, 1: seq1, 2: seq2, 3: both seq1 and seq2.
    seq_poly = [2, 1, 1, 0]

    # Used to indicates if state has changed and not yet saved.
    _dirty = False
    _last_saved = 0

    def __init__(self):
        super().__init__()

        # Configure EuroPi options to improve performance.
        b2.debounce_delay = 200
        oled.contrast(0)  # dim the oled

        # Current editable sequence.
        self.seq = self.seqs[0]

        self.page = 0
        self.param_index = 0
        self.counter = 0
        self.reset_timeout = 3000
        self._prev_k2 = None

        # Assign cv outputs to logical triggers.
        self.trigger_and = cv3
        self.trigger_xor = cv6

        # Save state struct
        self.format_string = "12s12s4s4s"
        self.State = namedtuple("State", "seq1 seq2 polys seq_poly")

        # Load state if previous state exists.
        self.load_state()

        @b1.handler
        def page_handler():
            # Pressing button 1 cycles through the pages of editable parameters.
            self._prev_k2 = None
            self.page = (self.page + 1) % len(self.pages)
            if self.page == 0:
                self.seq = self.seqs[0]
            if self.page == 1:
                self.seq = self.seqs[1]
            self._dirty = True

        @b2.handler
        def edit_parameter():
            # Pressing button 2 edits the current selected parameter.
            if self.page == 2:
                # Cycles through which sequence this polyrhythm is assigned to.
                self.seq_poly[self.param_index] = (
                    self.seq_poly[self.param_index] + 1) % len(self.seq_poly)
                self._dirty = True

        @din.handler
        def play_notes():
            # For each polyrhythm, check if each sequence is enabled and if the
            # current beat should play.
            seq1 = False
            seq2 = False
            # Check each polyrhythm to determine if a sequence should be triggered.
            for i, poly in enumerate(self.polys):
                if self.counter % poly == 0:
                    _seq1, _seq2 = self._trigger_seq(i)
                    seq1 = _seq1 or seq1
                    seq2 = _seq2 or seq2
            if seq1:
                self.seqs[0].play_next_step()
            if seq2:
                self.seqs[1].play_next_step()

            # Trigger logical AND / XOR
            if seq1 and seq2:
                self.trigger_and.on()
            if (seq1 or seq2) and seq1 != seq2:
                self.trigger_xor.on()

            self.counter = self.counter + 1

        @din.handler_falling
        def triggers_off():
            # Turn off all of the trigger CV outputs.
            for seq in self.seqs:
                seq.trigger_cv.off()
            self.trigger_and.off()
            self.trigger_xor.off()

    def _trigger_seq(self, step: int):
        # Convert poly sequence enablement into binary to determine which
        # sequences are triggered on this step.
        status = f"{self.seq_poly[step]:02b}"
        # Reverse the binary string values to match display.
        return int(status[1]) == 1, int(status[0]) == 1

    def load_state(self):
        """Load state from previous run."""
        state = self.load_state_bytes()
        if state:
            self.set_state(state)

    def save_state(self):
        """Save state if it has changed since last call."""
        # Only save state if state has changed and more than 1s has elapsed
        # since last save.
        if self._dirty and self.last_saved() > 1000:
            state = self.get_state()
            self.save_state_bytes(state)
            self._dirty = False
            self._last_saved = ticks_ms()

    def get_state(self):
        """Get state as a byte string."""
        return struct.pack(self.format_string,
                           self.seqs[0].get_state(),
                           self.seqs[1].get_state(),
                           bytes(self.polys),
                           bytes(self.seq_poly))

    def set_state(self, state):
        """Update instance variables with given state bytestring."""
        try:
            _state = self.State(*struct.unpack(self.format_string, state))
        except ValueError as e:
            print(f"Unable to load state: {e}")
            return
        self.seqs[0].set_state(_state.seq1)
        self.seqs[1].set_state(_state.seq2)
        self.polys = list(_state.polys)
        self.seq_poly = list(_state.seq_poly)

    def reset_check(self):
        """Reset the sequences and triggers when no clock pulse detected for specified time."""
        if self.counter != 0 and ticks_diff(ticks_ms(), din.last_triggered()) > self.reset_timeout:
            self.step = 0
            self.counter = 0
            for s in self.seqs:
                s.reset()
            self.trigger_and.off()
            self.trigger_xor.off()

    def show_menu_header(self):
        if ticks_diff(ticks_ms(), b1.last_pressed()) < MENU_DURATION:
            oled.fill_rect(0, 0, OLED_WIDTH, CHAR_HEIGHT, 1)
            oled.text(f"{self.pages[self.page]}", 0, 0, 0)

    def edit_sequence(self):
        # Display each sequence step.
        for step in range(len(self.seq.notes)):
            # If the current step is selected, edit with the parameter edit knob.
            if step == self.param_index:
                selected_note = k2.choice(NOTES)
                if self._prev_k2 and self._prev_k2 != selected_note:
                    self.seq.edit_step(step, selected_note)
                    self._dirty = True
                self._prev_k2 = selected_note

            # Display the current step.
            padding_x = 4 + (int(OLED_WIDTH/4) * step)
            padding_y = 12
            oled.text(f"{self.seq.notes[step]:<3}", padding_x, padding_y, 1)

            # Display a bar under current playing step.
            if step == self.seq.step_index:
                x1 = (int(OLED_WIDTH / 4) * step)
                x2 = int(OLED_WIDTH / 4)
                oled.fill_rect(x1, OLED_HEIGHT - 6, x2, OLED_HEIGHT, 1)

    def edit_poly(self):
        # Display each polyrhythm option.
        for poly_index in range(len(self.polys)):
            # If the current polyrhythm is selected, edit with the parameter knob.
            if poly_index == self.param_index:
                poly = k2.range(16) + 1
                if self._prev_k2 and self._prev_k2 != poly:
                    self.polys[poly_index] = poly
                    self._dirty = True
                self._prev_k2 = poly

            # Display the current polyrhythm.
            padding_x = 8 + (int(OLED_WIDTH/4) * poly_index)
            padding_y = 12
            oled.text(f"{self.polys[poly_index]:>2}", padding_x, padding_y, 1)

            # Display graphic for seq 1 & 2 enablement.
            seq1, seq2 = self._trigger_seq(poly_index)
            y1 = OLED_HEIGHT - 10

            x1 = 9 + int(OLED_WIDTH/4) * poly_index
            (oled.fill_rect if seq1 else oled.rect)(x1, y1, 6, 6, 1)

            x1 = 17 + int(OLED_WIDTH/4) * poly_index
            (oled.fill_rect if seq2 else oled.rect)(x1, y1, 6, 6, 1)

    def main(self):
        while True:
            oled.fill(0)

            # Parameter edit index & display selected box
            self.param_index = k1.range(4)
            left_x = int((OLED_WIDTH/4) * self.param_index)
            right_x = int(OLED_WIDTH/4)
            oled.rect(left_x, 0, right_x, OLED_HEIGHT, 1)

            if self.page == 0 or self.page == 1:
                self.edit_sequence()

            if self.page == 2:
                self.edit_poly()

            self.reset_check()

            self.show_menu_header()
            oled.show()

            self.save_state()


# Main script execution
if __name__ == '__main__':
    script = PolyrhythmSeq()
    script.main()



#!/usr/bin/env python3
"""
A loose clone of the ADDAC 508 "Swell Physics"

Uses a trochoidal wave to generate control voltages
"""

import configuration
from europi import *
from europi_script import EuroPiScript

from experimental.knobs import KnobBank
from experimental.math_extras import rescale
from experimental.screensaver import OledWithScreensaver
from experimental.thread import DigitalInputHelper

import machine
from math import (
    cos,
    pi,
    sin,
)

import _thread

two_pi = 2 * pi

ssoled = OledWithScreensaver()



class OceanSurge(EuroPiScript):
    BG_ERR = None

    MIN_RADIUS = 0.01
    MAX_RADIUS = 2

    MIN_LENGTH = 1
    MAX_LENGTH = 20

    MAX_BUOY_SPREAD = 10

    MIN_SPEED_INCREMENT = machine.freq() / 1_000_000_000_000
    MAX_SPEED_INCREMENT = MIN_SPEED_INCREMENT * 100

    CLIP_MODE_REFLECT = 0
    CLIP_MODE_WRAP = 1
    CLIP_MODE_CLIP = 2
    N_CLIP_MODES = 3

    CV_TARGET_NONE = -1
    CV_TARGET_AGITATION = 0
    CV_TARGET_BUOY_SPREAD = 1
    CV_TARGET_SPEED = 2
    CV_TARGET_SWELL_SIZE = 3

    def __init__(self):
        super().__init__()
        self.is_running = False

        self.wavelength = 10

        saved_state = self.load_state_json()
        self.speed = saved_state.get("speed", 0.5)
        self.spread = saved_state.get("spread", 0.5)
        self.swell_size = k1.percent()
        self.agitation = k2.percent()
        self.settings_dirty = False
        self.clip_mode = saved_state.get("clip_mode", self.CLIP_MODE_REFLECT)
        self.d = 2.0

        if self.config.CV_TARGET == "agitation":
            self.cv_target = self.CV_TARGET_AGITATION
        elif self.config.CV_TARGET == "buoy_spread":
            self.cv_target = self.CV_TARGET_BUOY_SPREAD
        elif self.config.CV_TARGET == "sim_speed":
            self.cv_target = self.CV_TARGET_SPEED
        elif self.config.CV_TARGET == "swell_size":
            self.cv_target = self.CV_TARGET_SWELL_SIZE
        else:
            self.cv_target = self.CV_TARGET_NONE

        self.shift = False

        self.k1_bank = (
            KnobBank.builder(k1)
            .with_unlocked_knob("swell_size")
            .with_locked_knob(
                "spread", initial_percentage_value=self.spread
            )
            .build()
        )

        self.k2_bank = (
            KnobBank.builder(k2)
            .with_unlocked_knob("agitation")
            .with_locked_knob(
                "speed", initial_percentage_value=self.speed
            )
            .build()
        )

        # Use B2 as a shift for knob controls
        def on_b2_press():
            self.shift = True
            self.k1_bank.set_current("spread")
            self.k2_bank.set_current("speed")
            ssoled.notify_user_interaction()

        def on_b2_release():
            self.shift = False
            self.k1_bank.set_current("swell_size")
            self.k2_bank.set_current("agitation")
            self.settings_dirty = True

        def on_b1_press():
            self.clip_mode = (self.clip_mode + 1) % self.N_CLIP_MODES
            self.settings_dirty = True
            ssoled.notify_user_interaction()

        self.digital_input_state = DigitalInputHelper(
            on_b2_rising = on_b2_press,
            on_b2_falling = on_b2_release,
            on_b1_rising = on_b1_press,
        )

    @classmethod
    def config_points(cls):
        return [
            configuration.choice(
                "CV_TARGET",
                [
                    "agitation",
                    "buoy_spread",
                    "sim_speed",
                    "swell_size",
                ],
                "agitation",
            ),
        ]

    def save(self):
        self.settings_dirty = False
        self.save_state_json({
            "spread": self.spread,
            "speed": self.speed,
        })

    def apply_clip(self, y):
        """
        Convert the wave to a CV value

        As configured, the wave will (at most) go from -2 to +2 on the Y axis. Depending on
        the clipping mode we either
            - truncate to -1 to +1
            - reflect at -1/+1
            - wrap through the limits (e.g. 1.5 -> -0.5)

        Then we shift the clipped wave to 0-1 & multiply by the max output voltage
        """
        if self.clip_mode == self.CLIP_MODE_CLIP:
            if y < -1:
                y = -1
            elif y > 1:
                y = 1
        elif self.clip_mode == self.CLIP_MODE_REFLECT:
            if y < -1 or y > 1:
                delta = abs(y) - 1
                if y > 1:
                    y = 1 - delta
                else:
                    y = -1 + delta
        elif self.clip_mode == self.CLIP_MODE_WRAP:
            if y < -1 or y > 1:
                delta = abs(y) - 1
                if y > 1:
                    y = -1 + delta
                else:
                    y = 1 - delta

        return y

    def wave_to_cv(self, y):
        y = (y + 1) / 2
        return y * MAX_OUTPUT_VOLTAGE

    def wave_x(self, a, b, t):
        return a + self.r * sin(t - 2 * pi * a / self.wavelength) * self.d ** b

    def wave_y(self, a, b, t):
        return b + self.r * cos(t - 2 * pi * a / self.wavelength) * self.d ** b

    def draw(self):
        # UI:
        # +----------------+
        # |Swl 0.0  Agt 0.0|
        # |Spr 0.0  Spd 0.0|
        # |                |
        # +----------------+
        ssoled.fill(0)


        if not self.shift:
            ssoled.fill_rect(0, 0, OLED_WIDTH, CHAR_HEIGHT+2, 1)
        else:
            ssoled.fill_rect(0, CHAR_HEIGHT+1, OLED_WIDTH, CHAR_HEIGHT+2, 1)

        ssoled.text(f"Swl {self.swell_size:0.1f}  Agt {self.agitation:0.1f}", 0, 1, 1 if self.shift else 0)
        ssoled.text(f"Spr {self.spread:0.1f}  Spd {self.speed:0.1f}", 0, CHAR_HEIGHT+2, 0 if self.shift else 1)
        if self.clip_mode == self.CLIP_MODE_CLIP:
            ssoled.text("clip", 0, 2*CHAR_HEIGHT, 1)
        elif self.clip_mode == self.CLIP_MODE_REFLECT:
            ssoled.text("reflect", 0, 2*CHAR_HEIGHT, 1)
        elif self.clip_mode == self.CLIP_MODE_WRAP:
            ssoled.text("wrap", 0, 2*CHAR_HEIGHT+2, 1)
        ssoled.show()

    def gui_thread(self):
        draw_rate = 30.0
        fps_sleep = 1.0 / draw_rate
        while self.is_running:
            try:
                self.draw()
                time.sleep(fps_sleep)
            except Exception as err:
                self.BG_ERR = err

        if self.is_running:
            self.BG_ERR = Exception('USB disconnected')

    def voltage_thread(self):
        sim_now = 0.0

        prev_swell = self.k1_bank["swell_size"].percent()
        prev_spread = self.k1_bank["spread"].percent()
        prev_agitation = self.k2_bank["agitation"].percent()
        prev_speed = self.k2_bank["speed"].percent()

        def ui_change(old, new):
            return abs(old - new) >= 0.01

        while self.is_running:
            if self.BG_ERR is not None:
                print(f'Background error {self.BG_ERR}')
                self.BG_ERR = None

            self.digital_input_state.update()

            # read the current knob values
            prev_swell = self.swell_size
            prev_spread = self.spread
            prev_agitation = self.agitation
            prev_speed = self.speed
            self.swell_size = self.k1_bank["swell_size"].percent()
            self.spread = self.k1_bank["spread"].percent()
            self.agitation = self.k2_bank["agitation"].percent()
            self.speed = self.k2_bank["speed"].percent()

            if (
                ui_change(prev_swell, self.swell_size)
                or ui_change(prev_spread, self.spread)
                or ui_change(prev_agitation, self.agitation)
                or ui_change(prev_speed, self.speed)
            ):
                ssoled.notify_user_interaction()

            if self.cv_target == self.CV_TARGET_AGITATION:
                self.agitation += ain.percent()
            elif self.cv_target == self.CV_TARGET_BUOY_SPREAD:
                self.spread += ain.percent()
            elif self.cv_target == self.CV_TARGET_SPEED:
                self.speed += ain.percent()
            elif self.cv_target == self.CV_TARGET_SWELL_SIZE:
                self.swell_size += ain.percent()

            # convert the knob values into our wave parameters
            self.r = rescale(self.agitation, 0.0, 1.0, self.MIN_RADIUS, self.MAX_RADIUS)
            self.wavelength = rescale(self.swell_size, 0.0, 1.0, self.MIN_LENGTH, self.MAX_LENGTH)

            buoy_x = self.MAX_BUOY_SPREAD * self.spread

            y1 = self.apply_clip(self.wave_y(-buoy_x, 0, sim_now))
            y2 = self.apply_clip(self.wave_y(0, 0, sim_now))
            y3 = self.apply_clip(self.wave_y(buoy_x, 0, sim_now))

            cv1.voltage(self.wave_to_cv(y1))
            cv2.voltage(self.wave_to_cv(y2))
            cv3.voltage(self.wave_to_cv(y3))

            if y1 < y2:
                cv4.on()
            else:
                cv4.off()

            if y2 > y3:
                cv5.on()
            else:
                cv5.off()

            cv6.voltage(
                self.wave_to_cv((y1 + y2 + y3) / 3.0)
            )

            sim_now += rescale(self.speed, 0, 1, self.MIN_SPEED_INCREMENT, self.MAX_SPEED_INCREMENT)
            if sim_now > two_pi:
                sim_now -= two_pi

            if self.settings_dirty:
                self.save()

    def main(self):
        self.is_running = True
        try:
            _thread.start_new_thread(self.gui_thread, ())
            self.voltage_thread()
        except KeyboardInterrupt as err:
            print(err)
            self.is_running = False
        finally:
            print("User aborted. Exiting.")

if __name__ == "__main__":
    OceanSurge().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
from time import sleep

from europi_script import EuroPiScript

try:
    # local dev
    from software.firmware.europi import (
        MAX_INPUT_VOLTAGE,
        OLED_HEIGHT,
        OLED_WIDTH,
        ain,
        b1,
        b2,
        cv1,
        cv2,
        cv4,
        din,
        k1,
        k2,
        oled,
    )
except ImportError:
    from europi import MAX_INPUT_VOLTAGE, OLED_HEIGHT, OLED_WIDTH, ain, b1, b2, cv1, cv2, cv4, din, k1, k2, oled

MAX_RATE = OLED_WIDTH
Y_TRUE = int(OLED_HEIGHT / 3 * 2)
Y_FALSE = int(OLED_HEIGHT / 3)
D_WAVE_HEIGHT = Y_TRUE - Y_FALSE + 1
RIGHT_EDGE = OLED_WIDTH - 1
Y_PIXELS = OLED_HEIGHT - 1


class Scope(EuroPiScript):
    """
    Can we make a functional oscilloscope on the EuroPi? Kinda! This script displays the inputs received on both the analog
    and digital inputs. Both waves can be toggled, and when no waves are displayed we switch to a text output, which also
    includes the knob values. The knobs each control an aspect of the x and y resolution respectively, allowing you to coax
    a little more detail out of the display.

    Note that the digital wave is displayed at a fixed y position that is not related to the value of the analog wave.

    Limitations include:
        - negative voltages are clipped to 0
        - screen height is only 32 pixels, so limited resolution
        - we can only refresh the screen so fast
        - analog pass through will be 'digitized'


    knob1 - samples per screen refresh. This effectively lets you 'zoom in' on the x-axis. Start at the lowest setting, 1 sample.
    knob2 - y voltage scale (only affects analog wave). This effectively lets you 'zoom in' on the low voltages. Start at
    the highest setting, 10v.

    button1 - toggle digital wave display
    button2 - toggle analog wave display

    out1 - digital in pass through
    out2 - analog in pass through
    out4 - inverted digital pass through
    """

    def __init__(self) -> None:
        super().__init__()
        self.enabled = [True, True]  # digital, analog

    def toggle(self, index):
        def f():
            self.enabled[index] = not self.enabled[index]

        return f

    @staticmethod
    def read_sample_rate():
        return k1.read_position(MAX_RATE) + 1

    @staticmethod
    def read_max_disp_voltage():
        return k2.read_position(math.ceil(MAX_INPUT_VOLTAGE)) + 1

    @staticmethod
    def calc_y_pos(max_disp_voltage, a_voltage):
        return Y_PIXELS - int(a_voltage / max_disp_voltage * Y_PIXELS)

    def main(self):
        b1.handler(self.toggle(0))
        b2.handler(self.toggle(1))

        oled.fill(0)
        old_value = din.value()

        while True:
            rate = self.read_sample_rate()
            max_disp_voltage = self.read_max_disp_voltage()

            for _ in range(rate):
                if any(self.enabled):
                    oled.scroll(-1, 0)
                    oled.vline(RIGHT_EDGE, 0, OLED_HEIGHT, 0)

                d_value = din.value()
                cv2.value(d_value)
                cv4.value(not d_value)

                if self.enabled[0]:  # digital wave
                    y_pos = Y_TRUE if d_value else Y_FALSE
                    oled.pixel(RIGHT_EDGE, OLED_HEIGHT - y_pos, 1)
                    if d_value != old_value:
                        oled.vline(RIGHT_EDGE, Y_FALSE + 1, D_WAVE_HEIGHT, 1)
                    old_value = d_value

                a_voltage = ain.read_voltage(1)
                cv1.voltage(a_voltage)

                if self.enabled[1]:  # analog wave
                    y_pos = self.calc_y_pos(max_disp_voltage, a_voltage)
                    oled.pixel(RIGHT_EDGE, y_pos, 1)

                if not self.enabled[0] and not self.enabled[1]:  # details output
                    rate = self.read_sample_rate()
                    max_disp_voltage = self.read_max_disp_voltage()
                    oled.fill(0)
                    oled.text(f"a: {a_voltage:4.1f}v d:{d_value}", 2, 3, 1)
                    oled.text(f"samples: {rate:3}", 2, 13, 1)
                    oled.text(f"y scale: {max_disp_voltage:4.1f}v", 2, 23, 1)
                    oled.show()

                sleep(0.001)

            oled.show()


if __name__ == "__main__":
    Scope().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
from time import ticks_diff, ticks_ms, sleep_ms
from random import randint, uniform
from europi_script import EuroPiScript
import machine
import json
import gc
import os
import micropython
import framebuf

'''
CVecorder
author: Nik Ansell (github.com/gamecat69)

digital_in: Clock input
analog_in: Incoming CV

button_1: Toggle recording stop/start
button_2: Short Press: Cycle through CV recorder channels. Long Press (0.5 seconds): Clear the current bank. Long Press (2 seconds): Clear all banks.

output_1: CV record / playback
output_2: CV record / playback
output_3: CV record / playback
output_4: CV record / playback
output_5: CV record / playback
output_6: CV record / playback

'''

'''
Ideas / to do:
- Add a morph capability using k2: left reduces CV values, right increases. Don’t adjust all values at once, so odds, then evens with each slight movement
'''

# Needed if using europi_script
class CVecorder(EuroPiScript):
    def __init__(self):

        # Needed if using europi_script
        super().__init__()

        # Micropython heap fragmentation notes:
        # - The pico has very limited memory and in some cases needs to be managed carefully
        # - In some cases you can get a MemoryError even if there is enough free memory, this is because micropython could not find enough contiguous memory because the heap has become fragmented
        # - Avoid heap fragmentation as much as possible by initializing and creating variables early, then updating. Rather than created new and destrying old variables.

        # Initialize variables
        self.step = 0
        self.stepLength = 64
        self.clockStep = 0
        self.ActiveCvr = 0
        self.ActiveBank = 0
        self.resetTimeout = 1000
        self.debug = False
        self.CvIn = 0
        self.bankToSave = 0
        self.initTest = False
        self.debugLogging = False
        self.errorString = ' '

        self.numCVR = 5  # Number of CV recorder channels - zero based
        self.numCVRBanks = 5  # Number of CV recording channel banks - zero based

        # Logging parameters
        self.logFilePrefix = 'cvecorder_debug'
        self.maxLogFiles = 5
        self.logFileList = []
        self.currentLogFile = ''
        self.maxLogFileName = self.logFilePrefix + str(self.maxLogFiles) + '.log'

        # rotate log files
        self.rotateLog()

        if self.debugLogging:
            self.writeToDebugLog(f"[init] Firing up!.")

        #self.CVR = []  # CV recorder channels
        #self.CvRecording = []  # CV recorder flags

        # Load CV Recordings from a previously stored state on disk or initialize if blank
        self.loadState()

        # Test routine, pick a random bank n times and save, then load the state
        if self.initTest:
            print(micropython.mem_info("level"))
            for n in range(3000):
                # Clear vars
                #self.CvRecording = []
                print(f"Running test: {n}")
                self.ActiveBank = randint(0, self.numCVRBanks)
                self.ActiveCvr = randint(0, self.numCVR)
                for i in range(0, self.stepLength-1):
                    self.CVR[self.ActiveBank][self.ActiveCvr][i] = uniform(0.0, 9.99)
                    #print(f"[{self.ActiveBank}][{self.ActiveCvr}][{i}] = {self.CVR[self.ActiveBank][self.ActiveCvr][i]}")
                self.bankToSave = self.ActiveBank
                self.saveState()
                self.loadState()

        @din.handler
        def dInput():
            self.handleClock()
            self.clockStep +=1

        @din.handler_falling
        def endClock():
            self.handleClock()

        @b1.handler
        def b1Pressed():
            # Set recording boolean to true and clear the recording buffer
            self.CvRecording[self.ActiveCvr] = 'pending'
            # Clear the array
            for n in range (0, self.stepLength):
                self.CVR[self.ActiveBank][self.ActiveCvr][n] = 0

        # # B2 Long press
        # @b2.handler_falling
        # def b2PressedLong():
        #     # Issue: This seems to get triggered randomly sometimes when the button is not pressed! This causes all CV banks to be cleared! :(
        #     # Leave this commented out until the problem is isolated and resolved
        #     # 2000ms press clears all banks
        #     if ticks_diff(ticks_ms(), b2.last_pressed()) >  2000:
        #         self.confirmDelete('all')
        #         self.clearCvrs('all')
        #         #self.bankToSave = self.ActiveBank
        #         #self.saveState()
        #         # reverse the ActiveCvr increment caused by the initial button press
        #         if self.ActiveCvr > 0:
        #             self.ActiveCvr -= 1
        #         else:
        #             self.ActiveCvr = self.numCVR
        #     # 500ms press clears the active bank
        #     elif ticks_diff(ticks_ms(), b2.last_pressed()) >  500:
        #         self.confirmDelete(self.ActiveBank)
        #         self.clearCvrs(self.ActiveBank)
        #         self.bankToSave = self.ActiveBank
        #         self.saveState()
        #         if self.debugLogging:
        #             self.writeToDebugLog(f"[b2PressedLong] > 500 Calling saveState() for bank {self.bankToSave}.")
        #         # reverse the ActiveCvr increment caused by the initial button press
        #         if self.ActiveCvr > 0:
        #             self.ActiveCvr -= 1
        #         else:
        #             self.ActiveCvr = self.numCVR

        # B2 short press
        @b2.handler
        def b2Pressed():
            # Change the active recorder channel
            if self.ActiveCvr < self.numCVR:
                self.ActiveCvr += 1
            else:
                self.ActiveCvr = 0

    def confirmDelete(self, bank):
        # Show confirm text on screen
        oled.fill(0)
        if str(bank) == 'all':
            oled.text('Clear ALL banks?', 0, 0, 1)
        else:
            oled.text(f'Clear bank {bank}?', 0, 0, 1)
        oled.text('CONFIRM: Hold B1', 0, 15, 1)
        oled.show()

        # Wait for button 1
        while b1.value() != 1:
            sleep_ms(250)

    def handleClock(self):

        # Sample input to 2 decimal places
        self.CvIn = round(ain.read_voltage(), 2)

        # Start recording if pending and on first step
        if self.step == 0 and self.CvRecording[self.ActiveCvr] == 'pending':
            self.CvRecording[self.ActiveCvr] = 'true'

        for i in range(self.numCVR+1):
            # If recording, write the sampled value to the CVR list and play the voltage
            if self.CvRecording[i] == 'true':
                self.CVR[self.ActiveBank][self.ActiveCvr][self.step] = self.CvIn
                cvs[self.ActiveCvr].voltage(self.CvIn)
            else:
                cvs[i].voltage(self.CVR[self.ActiveBank][i][self.step])

        # Reset step number at stepLength -1 as pattern arrays are zero-based
        if self.step < self.stepLength - 1:
            self.step += 1
        else:
            # Reset step to zero , stop recording and save recording to local storage
            self.step = 0
            if self.CvRecording[self.ActiveCvr] == 'true':
                self.CvRecording[self.ActiveCvr] = 'false'
                self.bankToSave = self.ActiveBank
                self.saveState()
                if self.debugLogging:
                    self.writeToDebugLog(f"[handleClock] Calling saveState() for bank {self.bankToSave}.")


    def clearCvrs(self, bank):
        for b in range(self.numCVRBanks+1):
            # skip bank unless 'all' is passed
            if b != bank and bank != 'all':
                continue
            if self.initTest:
                print('Clearing bank: ' + str(b))
            # Set all CV values to zero
            for i in range(self.numCVR+1):
                for n in range (0, self.stepLength):
                    self.CVR[b][i][n] = 0
            # Save the cleared bank to local storage
            self.bankToSave = b
            self.saveState()
            if self.debugLogging:
                self.writeToDebugLog(f"[clearCvrs] Calling saveState() for bank {self.bankToSave}.")

    # Currently not used, but keeping in this script for future use
    def initCvrs(self):
        for b in range(self.numCVRBanks+1):
            self.CVR.append([])
            for i in range(self.numCVR+1):
                self.CVR[b].append([])
                self.CvRecording.append('false')
                for n in range (0, self.stepLength):
                    self.CVR[b][i].append(0)

    def saveState(self):
        # generate output filename
        outputFile = f"saved_state_{self.__class__.__qualname__}_{self.bankToSave}.txt"

        # Convert each value to an int by multiplying by 100. This saves on storage and memory a little
        for i in range(len(self.CVR[self.bankToSave])):
            self.CVR[self.bankToSave][i] = [int(x * 100) for x in self.CVR[self.bankToSave][i]]

        if self.initTest:
            print('Saving state for bank: ' + str(self.bankToSave))

        # Trigger garbage collection to minimize memory use
        gc.collect()

        # Show free memory if running a debug test
        if self.initTest:
            print(self.free())

        # Write the value to a the state file
        maxRetries = 6
        attempts = 0
        while attempts < maxRetries:
            try:
                attempts += 1
                # Create json object of current CV bank
                jsonState = json.dumps(self.CVR[self.bankToSave])

                if self.debugLogging:
                    self.writeToDebugLog(f"[saveState] Saving state for bank: {str(self.bankToSave)}. Size: {len(jsonState)}")

                with open(outputFile, 'w') as file:
                    # Attempt write data to state on disk, then break from while loop if the return (num bytes written) > 0
                    if file.write(jsonState) > 0:
                        #self.errorString = ' '
                        if self.debugLogging:
                            self.writeToDebugLog(f"[saveState] Bank {str(self.bankToSave)} saved OK")
                        break
            except MemoryError as e:
                self.errorString = 'w'
                if self.initTest:
                    print(f'[{attempts}] Error: Memory allocation failed, retrying: {e}')
                if self.debugLogging:
                    self.writeToDebugLog(f"[saveState] Error: Memory allocation failed, retrying: {e}")
                    print(micropython.mem_info("level"))
                else:
                    pass

        # Convert from int back to float
        i=0
        for channel in self.CVR[self.bankToSave]:
            self.CVR[self.bankToSave][i] = [x / 100 for x in self.CVR[self.bankToSave][i]]
            i += 1

    def loadState(self):

        # For each bank, check if a state file exists, then load it
        # If not, initialize the bank with zeros then save it

        # Potential issue......
        # If for some reason the file open command fails, it will init each bank and wipe any previous recordings
        # Added retries and debug code to try and capture the error if this does occur

        self.CVR = []  # CV recorder channels
        self.CvRecording = []  # CV recorder flags

        # init cvRecording list
        for i in range(self.numCVR+1):
            self.CvRecording.append('false')

        for b in range(self.numCVRBanks+1):
            # Create a new array for the bank
            self.CVR.append([])

            # Check if a state file exists
            fileName = f"saved_state_{self.__class__.__qualname__}_{b}.txt"

            # Write the value to a the state file
            maxRetries = 2
            attempts = 0
            while attempts < maxRetries:

                try:
                    # save state exists for this bank, load it
                    with open(fileName, 'r') as file:

                        # read state from file into json object
                        jsonData = file.read()

                        if self.initTest:
                            print(f"Loading previous state for bank: {str(b)}. Size: {len(jsonData)}")

                        self.showLoadingScreen(str(b))
                        if self.debugLogging:
                            self.writeToDebugLog(f"[loadState] [{attempts}] Loading previous state for bank: {str(b)}. Size: {len(jsonData)}")

                        # populate CV recording channel with saved json data
                        self.CVR[b] = json.loads(jsonData)

                        # convert values in the list from int back to float
                        i=0
                        for channel in self.CVR[b]:
                            self.CVR[b][i] = [x / 100 if x > 0 else 0 for x in self.CVR[b][i]]
                            i += 1

                        # read OK, break from while loop
                        break

                except OSError as e:
                    self.errorString = 'r'
                    if self.debugLogging:
                        self.writeToDebugLog(f"[loadState] [{attempts}] No state file found for bank {b}. Error: {e}")
                    # No state file exists, initialize the array with zeros
                    if self.initTest:
                        print('Initializing bank: ' + str(b))

                    for i in range(self.numCVR+1):
                        self.CVR[b].append([])
                        for n in range (0, self.stepLength):
                            self.CVR[b][i].append(0)
                    # Save the state file for faster loading on next boot
                    self.bankToSave = b
                    self.saveState()

                except Exception as e:
                    self.errorString = 'x'
                    if self.debugLogging:
                        self.writeToDebugLog(f"[loadState] [{attempts}] Exception when attempting to open previous state file for bank {b}. {e}")

                # Sleep and increment attempt counter
                sleep_ms(50)
                attempts += 1

    # Currently not used, but keeping in this script for future use
    def debugDumpCvr(self):
        for b in range(self.numCVRBanks+1):
            for i in range(self.numCVR+1):
                print(str(b) + ':' + str(i) + ':' + str(self.CVR[b][i]))

    def free(self, full=False):
        #gc.collect()
        F = gc.mem_free()
        A = gc.mem_alloc()
        T = F+A
        P = '{0:.2f}%'.format(F/T*100)
        if not full: return P
        else : return ('Total:{0} Free:{1} ({2})'.format(T,F,P))

    def main(self):
        while True:
            self.getCvBank()
            self.updateScreen()

            # If I have been running, then stopped for longer than reset_timeout, reset the steps and clock_step to 0
            if self.clockStep != 0 and ticks_diff(ticks_ms(), din.last_triggered()) > self.resetTimeout:
                if self.CvRecording[self.ActiveCvr] != 'true':
                    self.step = 0
                    self.clockStep = 0

    def getCvBank(self):
        # Read CV Bank selection from knob 1
        self.ActiveBank = k1.read_position(self.numCVRBanks+1)

    # Rotate log files to avoid filling up storage
    def rotateLog(self):

        self.logFileList = os.listdir()

        # Delete the oldest allowed logfile if it exists
        if self.maxLogFileName in self.logFileList:
            os.remove(self.maxLogFileName)

        # Rename other log files if they exist 4 becomes 5 etc
        # Note: when this while loop exits self.currentLogFile is the name of the log file used by writeToDebugLog
        self.logFileNum = self.maxLogFiles - 1
        while self.logFileNum > 0:
            self.currentLogFile = self.logFilePrefix + str(self.logFileNum) + '.log'
            if self.currentLogFile in self.logFileList:
                os.rename(self.currentLogFile, self.logFilePrefix + str(self.logFileNum + 1) + '.log')
            self.logFileNum -= 1

    def writeToDebugLog(self, msg):

        try:
            rtc=machine.RTC()
            timestamp=rtc.datetime()
            timestring="%04d-%02d-%02d %02d:%02d:%02d"%(timestamp[0:3] + timestamp[4:7])
        except:
            timestring='0000-00-00 00:00:00'

        maxRetries = 6
        attempts = 0
        while attempts < maxRetries:
            try:
                attempts += 1
                with open(self.currentLogFile, 'a') as file:
                    # Attempt write data to state on disk, then break from while loop if the return (num bytes written) > 0
                    if file.write(timestring + ' ' + msg + '\n') > 0:
                        #self.errorString = ''
                        break
            except MemoryError as e:
                print(f'[{attempts}] Error: Memory allocation failed, retrying: {e}')
            except Exception as e:
                print(f'[{attempts}] Error writing to debug log. {e}')

    def showLoadingScreen(self, bank):
        # push the bytearray of the Rpi logo into a 32 x 32 framebuffer, then show on the screen

        buffer = bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00|?\x00\x01\x86@\x80\x01\x01\x80\x80\x01\x11\x88\x80\x01\x05\xa0\x80\x00\x83\xc1\x00\x00C\xe3\x00\x00~\xfc\x00\x00L'\x00\x00\x9c\x11\x00\x00\xbf\xfd\x00\x00\xe1\x87\x00\x01\xc1\x83\x80\x02A\x82@\x02A\x82@\x02\xc1\xc2@\x02\xf6>\xc0\x01\xfc=\x80\x01\x18\x18\x80\x01\x88\x10\x80\x00\x8c!\x00\x00\x87\xf1\x00\x00\x7f\xf6\x00\x008\x1c\x00\x00\x0c \x00\x00\x03\xc0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")
        fb = framebuf.FrameBuffer(buffer, 32, 32, framebuf.MONO_HLSB)
        oled.fill(0)
        oled.blit(fb, 0,0)
        oled.text(f'Loading ({bank})...', 40, 12, 1)
        oled.show()

    def updateScreen(self):
        # Clear the screen
        oled.fill(0)

        # Visualize each CV channel
        lPadding = 4
        # oled.fill_rect(x, y, width, height)
        oled.rect(lPadding+0 , 0, int(self.CVR[self.ActiveBank][0][self.step]*4), 11, 1)
        oled.rect(lPadding+42 , 0, int(self.CVR[self.ActiveBank][1][self.step]*4), 11, 1)
        oled.rect(lPadding+84 , 0, int(self.CVR[self.ActiveBank][2][self.step]*4), 11, 1)
        oled.rect(lPadding+0 , 12, int(self.CVR[self.ActiveBank][3][self.step]*4), 11, 1)
        oled.rect(lPadding+42 , 12, int(self.CVR[self.ActiveBank][4][self.step]*4), 11, 1)
        oled.rect(lPadding+84 , 12, int(self.CVR[self.ActiveBank][5][self.step]*4), 11, 1)

        # Show 'Rec' if recording
        if self.CvRecording[self.ActiveCvr] == 'true':
            oled.text('REC', 71, 25, 1)
        elif self.CvRecording[self.ActiveCvr] == 'pending':
            #oled.text('. .', 71, 25, 1)
            oled.text('.' + self.errorString + '.', 71, 25, 1)
        else:
            oled.text(' ' + self.errorString + ' ', 71, 25, 1)

        # Active recording channel
        oled.text(str(self.ActiveBank+1) + ':' + str(self.ActiveCvr+1), 100, 25, 1)

        # Current step
        oled.rect(lPadding-1, 26, 64, 6, 1)
        oled.fill_rect(lPadding-1, 26, self.step, 6, 1)

        oled.show()

if __name__ == '__main__':
    dm = CVecorder()
    dm.main()



#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Lutra -- A EuroPi reimagining of Expert Sleepers' Otterley

This script is a free-running, syncable LFO with variable shapes.  Each output channel's clock speed is slightly
different, with the spread being controlled by K2 (and optionally AIN).  The overall speed is controlled by
K1 (and optionally AIN).  B1 & DIN both act as synchronization inputs. B2 cycles through the different output
waveforms.

Lutra: a genus of otters, including L. lutra, the eurasian otter and L. sumatrana, the hairy-nosed otter.

@author Chris Iverach-Brereton
@year   2024

@see    https://expert-sleepers.co.uk/otterley.html
"""

from europi import *
from europi_script import EuroPiScript

import configuration
import math
from random import random
import time

import _thread

from experimental.math_extras import rescale
from experimental.thread import DigitalInputHelper

class WaveGenerator:
    """Generates the output wave forms and sets the voltage going to one of the output jacks

    5 wave shapes are supported, with the cycle time expressed in "ticks." These ticks have no 1:1 relationship
    with any real-world time unit, and are simply defined by each iteration through the script's main loop.
    """

    ## Supported wave shapes
    WAVE_SHAPE_SINE = 0
    WAVE_SHAPE_SQUARE = 1
    WAVE_SHAPE_TRIANGLE = 2
    WAVE_SHAPE_SAW = 3
    WAVE_SHAPE_RAMP = 4
    WAVE_SHAPE_STEP_RANDOM = 5
    WAVE_SHAPE_SMOOTH_RANDOM = 6
    NUM_WAVE_SHAPES = 7

    WAVE_SHAPES_NAMES = [
        "Sine",
        "Square",
        "Triangle",
        "Saw",
        "Ramp",
        "Step_Random",
        "Smooth_Random",
    ]

    WAVE_SHAPES_NAMES_TO_SHAPES = {
        "sine": WAVE_SHAPE_SINE,
        "square": WAVE_SHAPE_SQUARE,
        "triangle": WAVE_SHAPE_TRIANGLE,
        "saw": WAVE_SHAPE_SAW,
        "ramp": WAVE_SHAPE_RAMP,
        "step_random": WAVE_SHAPE_STEP_RANDOM,
        "smooth_random": WAVE_SHAPE_SMOOTH_RANDOM,
    }

    ## 12x12 pixel images of the wave shapes
    WAVE_SHAPE_IMAGES = [
        FrameBuffer(bytearray(b'\x10\x00(\x00D\x00D\x00\x82\x00\x82\x00\x82\x10\x82\x10\x01\x10\x01\x10\x00\xa0\x00@'), 12, 12, MONO_HLSB),              # SINE
        FrameBuffer(bytearray(b'\xfe\x00\x82\x00\x82\x00\x82\x00\x82\x00\x82\x00\x82\x00\x82\x00\x82\x00\x82\x00\x82\x00\x83\xf0'), 12, 12, MONO_HLSB),  # SQUARE
        FrameBuffer(bytearray(b'\x06\x00\x06\x00\t\x00\t\x00\x10\x80\x10\x80 @ @@ @ \x80\x10\x80\x10'), 12, 12, MONO_HLSB),                              # TRIANGLE
        FrameBuffer(bytearray(b'\x80\x00\xc0\x00\xa0\x00\x90\x00\x88\x00\x84\x00\x82\x00\x81\x00\x80\x80\x80@\x80 \x80\x10'), 12, 12, MONO_HLSB),        # SAW
        FrameBuffer(bytearray(b'\x00\x10\x000\x00P\x00\x90\x01\x10\x02\x10\x04\x10\x08\x10\x10\x10 \x10@\x10\x80\x10'), 12, 12, MONO_HLSB),              # RAMP
        FrameBuffer(bytearray(b'\x00\xe0\x00\xa0\x00\xa0\x00\xa0<\xa0$\xa0$\xa0\xe4\xb0\x04\x80\x04\x80\x04\x80\x07\x80'), 12, 12, MONO_HLSB),           # STEP_RANDOM
        FrameBuffer(bytearray(b'\x00`\x00P\x00P\x08P\x14P$PD\x90\x82\x80\x02\x80\x02\x80\x02\x80\x01\x80'), 12, 12, MONO_HLSB),                          # SMOOTH_RANDOM
    ]

    IMAGE_SIZE = (12, 12)

    def __init__(self, cv_output):
        """Constructor

        @param cv_output  The CV output jack that the generated wave gets output on
        """
        self.cv_output = cv_output
        self.shape = 0

        self.cycle_ticks = 1000
        self.current_tick = 0

        self.prev_random_goal = random() * MAX_OUTPUT_VOLTAGE
        self.random_goal = random() * MAX_OUTPUT_VOLTAGE

    def reset(self):
        """Reset the wave to the beginning
        """
        self.current_tick = 0

    def change_cycle_length(self, new_length):
        """Change the number of steps in the pattern

        We need to preserve our relative progress to avoid skipping when changing the cycle length
        """
        if new_length != self.cycle_ticks:
            progress = self.current_tick / self.cycle_ticks
            self.cycle_ticks = new_length
            self.current_tick = int(new_length * progress)

    def tick(self):
        """Calculate the appropriate voltage for the output, given the current clock time

        @return The desired voltage
        """
        if self.shape == self.WAVE_SHAPE_SINE:
            # we want to start at -1 and go up, so we actually use a negative cos wave, but the shape is the same
            theta = (self.current_tick / self.cycle_ticks) * 2 * math.pi
            volts = (-math.cos(theta) + 1) / 2 * MAX_OUTPUT_VOLTAGE
        elif self.shape == self.WAVE_SHAPE_SQUARE:
            if self.current_tick < (self.cycle_ticks >> 1):
                volts = MAX_OUTPUT_VOLTAGE
            else:
                volts = 0
        elif self.shape == self.WAVE_SHAPE_TRIANGLE:
            half_cycle_ticks = (self.cycle_ticks >> 1)
            if self.current_tick < half_cycle_ticks:
                volts = self.current_tick / half_cycle_ticks * MAX_OUTPUT_VOLTAGE
            else:
                volts = MAX_OUTPUT_VOLTAGE - (self.current_tick - half_cycle_ticks) / half_cycle_ticks * MAX_OUTPUT_VOLTAGE
        elif self.shape == self.WAVE_SHAPE_SAW:
            volts = MAX_OUTPUT_VOLTAGE - self.current_tick / self.cycle_ticks * MAX_OUTPUT_VOLTAGE
        elif self.shape == self.WAVE_SHAPE_RAMP:
            volts = self.current_tick / self.cycle_ticks * MAX_OUTPUT_VOLTAGE
        elif self.shape == self.WAVE_SHAPE_STEP_RANDOM:
            volts = self.random_goal
        elif self.shape == self.WAVE_SHAPE_SMOOTH_RANDOM:
            slope = (self.random_goal - self.prev_random_goal) / self.cycle_ticks
            volts = slope * self.current_tick + self.prev_random_goal
        else:
            volts = 0

        self.current_tick = self.current_tick + 1
        if self.current_tick >= self.cycle_ticks:
            self.current_tick = 0
            self.prev_random_goal = self.random_goal
            self.random_goal = random() * MAX_OUTPUT_VOLTAGE

        self.cv_output.voltage(volts)
        return volts

class Lutra(EuroPiScript):
    """The main class for this script; handles running the main loop, configuring I/O, loading, and saving state.
    """

    # We support CV control over either LFO speed OR LFO spread via AIN.  This option is not exposed through a menu
    # and must be configured via the config file.  See lutra.md for details
    AIN_MODE_SPREAD = 0
    AIN_MODE_SPEED = 1
    AIN_MODE_NAMES = [
        "Spread",
        "Speed"
    ]

    # The maximum and minimum cycle time for the LFOs
    MIN_CYCLE_TICKS = 250
    MAX_CYCLE_TICKS = 10000

    # Maximum wave speed multipliers relative to cv1
    MAX_SPEED_MULTIPLIERS = [
        1/1,
        6/5,
        5/4,
        4/3,
        3/2,
        2/1
    ]

    def __init__(self):
        """Constructor

        This creates all of the necessary objects, but does not create the separate thread for handling the GUI;
        see @main for that.
        """
        super().__init__()

        self.waves = [
            WaveGenerator(cv) for cv in cvs
        ]
        self.config_dirty = False
        self.hold_low = False
        self.last_wave_change_at = time.ticks_ms()

        # Connect the B2 handler to our digital input helper
        # B1 and DIN are handled differently since they interact with each other
        # See @wave_generator_thread
        self.digital_input_state = DigitalInputHelper(
            on_b2_rising = self.on_b2_rising
        )

        # Save the last screen width's worth of output voltages converted to pixel heights
        # This speeds up rendering
        self.display_pixels = [
            [] for cv in cvs
        ]

        # To coordinate access to self.display_pixels between threads we need a mutex to make sure
        # we don't read the array while it's being modified
        self.pixel_lock = _thread.allocate_lock()

        self.load()

    @classmethod
    def config_points(cls):
        """Return the static configuration options for this class
        """
        return [
            configuration.choice(name="AIN_MODE", choices=["spread", "speed"], default="spread")
        ]

    def load(self):
        """Load and apply the saved state

        @exception  ValueError if the configuration contains invalid values
        """
        cfg = self.load_state_json()
        shape = cfg.get("wave", "sine").lower()
        if shape in WaveGenerator.WAVE_SHAPES_NAMES_TO_SHAPES.keys():
            for wave in self.waves:
                wave.shape = WaveGenerator.WAVE_SHAPES_NAMES_TO_SHAPES[shape]
        else:
            raise ValueError(f"Unknown wave shape: {shape}")

    def save(self):
        """Write the saved-state file & set config_dirty to False
        """
        cfg = {
            "wave": WaveGenerator.WAVE_SHAPES_NAMES[self.waves[0].shape].lower()
        }
        self.save_state_json(cfg)
        self.config_dirty = False

    def on_digital_in_rising(self):
        """Called when either B1 or DIN goes high

        Signals the wave generator thread that all outputs should be forced low
        """
        self.hold_low = True

    def on_digital_in_falling(self):
        """Called when both B1 and DIN are low

        Signals the wave generator thread that all outputs should reset
        """
        self.hold_low = False
        for wave in self.waves:
                wave.reset()

    def on_b2_rising(self):
        """Called when either B2 goes high

        Cycles through the active wave shape
        """
        shape = (self.waves[0].shape + 1) % WaveGenerator.NUM_WAVE_SHAPES
        for wave in self.waves:
            wave.shape = shape
        self.last_wave_change_at = time.ticks_ms()
        self.config_dirty = True

    def gui_render_thread(self):
        """A thread function that handles drawing the GUI
        """
        SHOW_WAVE_TIMEOUT = 3000

        while self.is_running:
            now = time.ticks_ms()
            oled.fill(0)
            with self.pixel_lock:
                for channel in self.display_pixels:
                    for px in range(len(channel)):
                        oled.pixel(px, channel[px], 1)
            if time.ticks_diff(now, self.last_wave_change_at) < SHOW_WAVE_TIMEOUT:
                oled.blit(WaveGenerator.WAVE_SHAPE_IMAGES[self.waves[0].shape], 0, 0)
            oled.show()

    def wave_generation_thread(self):
        """A thread function that handles the underlying math of generating the waveforms
        """
        while self.is_running:
            # Read the digital inputs
            self.digital_input_state.update()

            # Manually handle B1 and DIN rising & falling
            # If either goes high, signal that we want to old the outputs low
            # If both become low, signal that all outputs should reset & output normally
            if self.digital_input_state.b1_rising or self.digital_input_state.din_rising:
                self.on_digital_in_rising()
            elif (
                (self.digital_input_state.b1_falling and not self.digital_input_state.din_high) or
                (self.digital_input_state.din_falling and not self.digital_input_state.b1_pressed)
            ):
                self.on_digital_in_falling()

            # Read the CV inputs and apply them
            # Round to 2 decimal places to reduce noise
            ain_percent = round(ain.percent(), 2)
            k1_percent = round(k1.percent(), 2)
            k2_percent = round(k2.percent(), 2)

            if self.config.AIN_MODE == self.AIN_MODE_SPREAD:
                k_speed = k1_percent
                k_spread = clamp(k2_percent + ain_percent, 0, 1)
            else:
                k_speed = clamp(k1_percent + ain_percent, 0, 1)
                k_spread = k2_percent

            base_ticks = int((1.0 - k_speed) * (self.MAX_CYCLE_TICKS - self.MIN_CYCLE_TICKS) + self.MIN_CYCLE_TICKS)
            for i in range(len(cvs)):
                base_tick_multiplier = rescale(k_spread, 0, 1, 1, self.MAX_SPEED_MULTIPLIERS[i])
                spread_ticks = int(base_ticks / base_tick_multiplier)
                self.waves[i].change_cycle_length(spread_ticks)

            for i in range(len(cvs)):
                pixel_height = OLED_HEIGHT - 1
                if self.hold_low:
                    cvs[i].off()
                else:
                    volts = self.waves[i].tick()
                    pixel_height = int(OLED_HEIGHT - 1 - volts / MAX_OUTPUT_VOLTAGE * (OLED_HEIGHT-1))

                with self.pixel_lock:
                    self.display_pixels[i].append(pixel_height)
                    if len(self.display_pixels[i]) >= OLED_WIDTH:
                        self.display_pixels[i].pop(0)

            if self.config_dirty:
                self.save()

    def main(self):
        self.is_running = True
        try:
            _thread.start_new_thread(self.gui_render_thread, ())
            self.wave_generation_thread()
        except KeyboardInterrupt:
            self.is_running = False
        finally:
            print("User aborted. Exiting.")

if __name__ == "__main__":
    Lutra().main()



# Copyright 2023 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Custom font demo
author: François Georgy (https://github.com/francoisgeorgy)
date: 2023-03-30

Use this script to test the support for custom fonts.

Please refer to experimental/custom_font.md documentation for more information.
"""
import time

from europi_script import EuroPiScript

from experimental import custom_font
from experimental import freesans14
from experimental import freesans17
from experimental import freesans20
from experimental import freesans24

oled = custom_font.oled


class CustomFontDemo(EuroPiScript):

    @classmethod
    def display_name(cls):
        return "Custom font demo"

    def __init__(self):
        super().__init__()
        self.demo = 0

    def update_demo(self):
        self.demo = (self.demo + 1) % 11
        if self.demo == 0:
            oled.centre_text("size 8")
        elif self.demo == 1:
            oled.centre_text("size 14", font=freesans14)
        elif self.demo == 2:
            oled.centre_text("size 17", font=freesans17)
        elif self.demo == 3:
            oled.centre_text("size 20", font=freesans20)
        elif self.demo == 4:
            oled.centre_text("size 24", font=freesans24)
        elif self.demo == 5:
            oled.fill(0)
            oled.text("8", 2, 20)
            oled.text("14", 18, 16, font=freesans14)
            oled.text("17", 40, 13, font=freesans17)
            oled.text("20", 66, 10, font=freesans20)
            oled.text("24", 96, 6, font=freesans24)
            oled.show()
        elif self.demo == 6:
            oled.centre_text("Default\n8x8 font\ncentered")
        elif self.demo == 7:
            oled.centre_text("FreeSans 14\nline 2", font=freesans14)
        elif self.demo == 8:
            oled.centre_text("FreeSans 17", font=freesans17)
        elif self.demo == 9:
            oled.centre_text("FreeSans 24", font=freesans24)
        elif self.demo == 10:
            oled.fill(0)
            oled.text("Fonts can", 0, 0)
            oled.text("be", 0, 10, font=freesans24)
            oled.text("mixed", 35, 14, font=freesans14)
            oled.text("freely", 80, 10, font=freesans20)
            oled.show()

    def main(self):
        t = 0
        while True:
            if (time.time() - t) >= 1:
                self.update_demo()
                t = time.time()


if __name__ == "__main__":
    oled.contrast(0)  # dim the display
    CustomFontDemo().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
from europi_script import EuroPiScript


class BinaryCounter(EuroPiScript):
    MAX_N = (1 << NUM_CVS) - 1

    def __init__(self):
        super().__init__()

        self.n = 0
        self.k = int(
            (
                k1.percent() + k2.percent() * ain.percent()
            ) * self.MAX_N
        )

        self.gate_recvd = False

        din.handler(self.on_gate_rise)
        din.handler_falling(self.on_gate_fall)
        b1.handler(self.on_gate_rise)
        b1.handler_falling(self.on_gate_fall)

        b2.handler(self.reset)

    def on_gate_rise(self):
        self.gate_recvd = True

    def on_gate_fall(self):
        turn_off_all_cvs()

    def reset(self):
        self.n = 0

    def set_outputs(self):
        for i in range(NUM_CVS):
            if (self.n >> i) & 0x01:
                cvs[i].on()
            else:
                cvs[i].off()

    def main(self):
        while True:
            self.k = int(
                (
                    k1.percent() + k2.percent() * ain.percent()
                ) * self.MAX_N
            )

            if self.gate_recvd:
                self.set_outputs()
                self.n = (self.n + self.k) & self.MAX_N
                self.gate_recvd = False

            oled.centre_text(f"""k = {self.k}
{self.n:06b}""")


if __name__ == "__main__":
    BinaryCounter().main()



# Copyright 2022 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
import machine
from time import ticks_diff, ticks_ms
from random import randint, uniform, choice

from europi_script import EuroPiScript

'''
Hamlet
author: Sean Bechhofer (github.com/seanbechhofer)
date: 2022-04-16
labels: sequencer, triggers, drums, randomness

A gate and CV sequencer that builds on Nik Ansell's Consequencer. Changes are:
* Slimmed down drum patterns to two instruments, BD/HH
* Two channels of gate and random stepped CV with sparsity control. As
  the sparsity is turned up, notes will drop out of the pattern.

digital_in: clock in
analog_in: sparsity CV

knob_1: sparsity
knob_2: select pre-loaded drum pattern

button_1: Short Press: Play previous CV Pattern. Long Press: Change CV track length multiplier
button_2: Short Press: Generate a new random cv pattern for Tracks 1 and 2. Long Press: Cycle through analogue input modes
output_1: trigger 1 / Bass Drum
output_2: trigger 2 / Hi-Hat
output_3: gates for track 1
output_4: gates for track 2
output_5: randomly generated track 2 CV (cycled by pushing button 2)
output_6: randomly generated track 1 CV (cycled by pushing button 2)

'''

class Hamlet(EuroPiScript):
    def __init__(self):
        self.bd = cv1
        self.hh = cv2
        self.gate_1 = cv3
        self.cv_1 = cv6
        self.gate_2 = cv4
        self.cv_2 = cv5

        # Initialize sequencer pattern arrays
        p = pattern()
        self.BD=p.BD
        self.HH=p.HH

        # Initialize variables
        self.drum_step = 0
        self.track_step = 0

        self.trigger_duration_ms = 50
        self.clock_step = 0
        self.pattern = 0
        self.minAnalogInputVoltage = 0.9
        self.randomness = 0
        self.sparsity = 0
        self.analogInputMode = 1 # 1: Randomness, 2: Pattern, 3: CV Pattern
        self.CvPattern = 0

        # Generate random CV for cv4-6
        self.track_1 = []
        self.track_2 = []

        self.track_multiplier = 1 # multiplies CV pattern length so we can have 16, 32, 64

        self.generateNewRandomCVPattern()

        # Triggered when button 2 is released.
        # Short press: Generate random CV for Tracks 1 and 2
        # Long press: Change operating mode
        @b2.handler_falling
        def b2Pressed():

            if ticks_diff(ticks_ms(), b2.last_pressed()) >  300:
                if self.analogInputMode < 3:
                    self.analogInputMode += 1
                else:
                    self.analogInputMode = 1
            else:
                # Move to next cv pattern if one already exists, otherwise create a new one
                self.CvPattern += 1
                if self.CvPattern == len(self.track_1):
                    self.generateNewRandomCVPattern()

        # Triggered when button 1 is released
        # Short press: Play previous CV pattern for cv4-6
        # Long press: Not used
        @b1.handler_falling
        def b1Pressed():
            if ticks_diff(ticks_ms(), b1.last_pressed()) >  300:
                if self.track_multiplier >=4:
                    self.track_multiplier = 1
                else:
                    self.track_multiplier *= 2
                pass
            else:
                # Play previous CV Pattern, unless we are at the first pattern
                if self.CvPattern != 0:
                    self.CvPattern -= 1

        # Triggered on each clock into digital input. Output triggers.
        @din.handler
        def clockTrigger():

            self.step_length = len(self.BD[self.pattern])

            # As the randomness value gets higher, the chance of a randomly selected int being lower gets higher
            if randint(0,99) < self.randomness:
                self.bd.value(randint(0, 1))
                self.bd.value(randint(0, 1))
            else:
                # Trigger drums
                self.bd.value(int(self.BD[self.pattern][self.drum_step]))
                self.hh.value(int(self.HH[self.pattern][self.drum_step]))

            # A pattern was selected which is shorter than the current step. Set to zero to avoid an error
            if self.drum_step >= self.step_length:
                self.drum_step = 0

            track_1_cv, track_1_sparsity = self.track_1[self.CvPattern][self.track_step]
            track_2_cv, track_2_sparsity = self.track_2[self.CvPattern][self.track_step]

            # Set track 1 and 2 voltage outputs based on previously
            # generated random pattern and sparsity. CV only changed
            # if sparsity value is above the current control level. So
            # as sparsity increases, notes decrease.

            if track_1_sparsity > self.sparsity:
                self.cv_1.voltage(track_1_cv)
                self.gate_1.on()
            else:
                self.gate_1.off()

            if track_2_sparsity > self.sparsity:
                self.cv_2.voltage(track_2_cv)
                self.gate_2.on()
            else:
                self.gate_2.off()

            # Reset clock step at 128 to avoid a HUGE integer if running for a long time
            # over a really long period of time this would look like a memory leak
            if self.clock_step < 128:
                self.clock_step +=1
            else:
                self.clock_step = 0

            # Reset step number at step_length -1 as pattern arrays are zero-based
            if self.drum_step < self.step_length - 1:
                self.drum_step += 1
            else:
                self.drum_step = 0

            if self.track_step < (self.step_length * self.track_multiplier) - 1:
                self.track_step += 1
            else:
                self.track_step = 0


        @din.handler_falling
        def clockTriggerEnd():
            self.bd.off()
            self.hh.off()
            self.gate_1.off()
            self.gate_2.off()

    def generateNewRandomCVPattern(self):
        """Generate new random CV patterns for the voice tracks"""
        self.step_length = len(self.BD[self.pattern])
        # CV patterns are up to 4 times the length of the drum pattern
        patt = (self.generateRandomPattern(self.step_length, 0, 9) +
                  self.generateRandomPattern(self.step_length, 0, 9) +
                  self.generateRandomPattern(self.step_length, 0, 9) +
                  self.generateRandomPattern(self.step_length, 0, 9))
        self.track_1.append(patt)
        patt = (self.generateRandomPattern(self.step_length, 0, 9) +
                  self.generateRandomPattern(self.step_length, 0, 9) +
                  self.generateRandomPattern(self.step_length, 0, 9) +
                  self.generateRandomPattern(self.step_length, 0, 9))
        self.track_2.append(patt)

    def updatePattern(self):
        """Read from knob 2 or CV (if in appropriate mode) and change the drum pattern accordingly"""
        # If mode 2 and there is CV on the analogue input use it, if
        # not use the knob position
        val = 100 * ain.percent()
        if self.analogInputMode == 2 and val > self.minAnalogInputVoltage:
            self.pattern = int((len(self.BD) / 100) * val)
        else:
            self.pattern = k2.read_position(len(self.BD))

        self.step_length = len(self.BD[self.pattern])

    def updateCvPattern(self):
        """Read from CV (if in appropriate mode) and change the CV pattern accordingly"""
        # If analogue input mode 3, get the CV pattern from CV input
        if self.analogInputMode != 3:
            return
        else:
            # Get the analogue input voltage as a percentage
            CvpVal = 100 * ain.percent()

            # Is there a voltage on the analogue input and are we configured to use it?
            if CvpVal > 0.4:
                # Convert percentage value to a representative index of the pattern array
                self.CvPattern = int((len(self.track_1) / 100) * CvpVal)


    def generateRandomPattern(self, length, min, max):
        """Generate a new random pattern"""
        # Returns a list of pairs of value and a sparsity
        self.t=[]
        # Sparsity values are from 1 to length. This means that notes
        # will progressively drop out as sparsity increases. Using
        # random values would give different behaviour.
        # No shuffle in micropython random :-(
        # sparsities = random.shuffle(range(1,length+1))
        count = 0
        sparsities = []
        numbers = list(range(1,length+1))
        while len(numbers) > 0:
                chosen = choice(numbers)
                sparsities.append(chosen)
                numbers.remove(chosen)

        for i in range(0, length):
            self.t.append((uniform(0,9),sparsities[i]))
        return self.t

    def updateSparsity(self):
        """Update sparsity value from knob 1"""
        # Don't use Analog input for now
        self.sparsity = k1.read_position(steps=len(self.BD[self.pattern])+1)

    def updateRandomness(self):
        """Read randomness from CV (if in appropriate mode)"""
        # If mode 1 and there is CV on the analogue input use it
        val = 100 * ain.percent()
        if self.analogInputMode == 1 and val > self.minAnalogInputVoltage:
            self.randomness = val

    def main(self):
        while True:
            self.updatePattern()
            self.updateSparsity()
            self.updateRandomness()
            self.updateCvPattern()
            self.updateScreen()
            self.reset_timeout = 500
            # If I have been running, then stopped for longer than
            # reset_timeout, reset the steps and clock_step to 0
            if self.clock_step != 0 and ticks_diff(ticks_ms(), din.last_triggered()) > self.reset_timeout:
                self.drum_step = 0
                self.clock_step = 0

    def visualizePattern(self, pattern):
        self.t = pattern
        self.t = self.t.replace('1','o')
        self.t = self.t.replace('0',' ')
        return self.t

    def visualizeTrack(self, track):
        t = ''
        for i in range(0, len(track)):
            track_cv, track_sparsity = track[i]
            if track_sparsity > self.sparsity:
                t += '.'
            else:
                t += ' '
        return t

    def updateScreen(self):
        oled.fill(0)

        # Show selected pattern visually
        oled.text(self.visualizePattern(self.BD[self.pattern]),0,0,1)
        oled.text(self.visualizePattern(self.HH[self.pattern]),0,8,1)
        oled.text(self.visualizeTrack(self.track_1[self.CvPattern]),0,16,1)

        # Show the analogInputMode
        oled.text('M' + str(self.analogInputMode), 112, 25, 1)

        # Show sparsity
        oled.text('S' + str(int(self.sparsity)), 32, 25, 1)

        # Show CV pattern
        oled.text('C' + str(self.CvPattern), 60, 25, 1)

        # Show CV pattern
        oled.text('x' + str(self.track_multiplier), 80, 25, 1)

        # Show randomness
        oled.text('R' + str(int(self.randomness)), 4, 25, 1)

        oled.show()

class pattern:

    # Initialize pattern lists
    BD=[]
    HH=[]

     # Add patterns. This is a smaller number of patterns than in the
    # original Consequencer.

    BD.append("0000000000000000")
    HH.append("0000000000000000")

    BD.append("1000100010001000")
    HH.append("0000000000000000")

    BD.append("1000100010001000")
    HH.append("0010001000100010")

    BD.append("1000100010001000")
    HH.append("0010001000100011")

    BD.append("1000100010001000")
    HH.append("0111011101110111")

    BD.append("1000100010001000")
    HH.append("1111111111111111")

    BD.append("0000000000000000")
    HH.append("0010001000100010")

    BD.append("1000100010001000")
    HH.append("0000000001111111")

    BD.append("0000000000000000")
    HH.append("1111111011111110")

    BD.append("1000100010010100")
    HH.append("1111111011101110")

    BD.append("1000001100100000")
    HH.append("1111111111111111")

    BD.append("1000100010001000")
    HH.append("1111101111111101")

    BD.append("1000100010010100")
    HH.append("0010001000100011")

    BD.append("1000100010010010")
    HH.append("0000000000000000")

    BD.append("1000100010010010")
    HH.append("0010001000100010")

if __name__ == '__main__':
    # Reset module display state.
    oled.fill(0)
    turn_off_all_cvs()
    hm = Hamlet()
    hm.main()



#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Generates CV and gate signals based on the slope of an incoming analogue signal
"""

from europi import *
from europi_script import EuroPiScript

from experimental.math_extras import median

class Slopes(EuroPiScript):
    noise_attenuator = k1
    output_attenuator = k2

    ain_rising_gate = cv1
    ain_falling_gate = cv2
    ain_steady_gate = cv3
    ain_pos_slope_cv = cv4
    ain_neg_slope_cv = cv5
    ain_slope_magnitude_cv = cv6

    MIN_BINS = 1
    MAX_BINS = 12

    MIN_SAMPLES = 32
    MAX_SAMPLES = 512

    MAX_SLOPE = europi_config.MAX_INPUT_VOLTAGE / 10.0

    DEAD_ZONE = 0.01

    def __init__(self):
        super().__init__()

        # bins used for median smoothing of the raw AIN data
        self.bins = [0] * self.MAX_BINS

    def denoise(self, window_size):
        return median(self.bins[len(self.bins) - window_size - 1:])

    def main(self):
        turn_off_all_cvs()
        oled.fill(0)
        oled.show()

        while True:
            denoise_percent = self.noise_attenuator.percent()
            n_bins = round(denoise_percent * (self.MAX_BINS - self.MIN_BINS) + self.MIN_BINS)
            n_samples = round(denoise_percent * (self.MAX_SAMPLES - self.MIN_SAMPLES) + self.MIN_SAMPLES)
            deadzone = self.DEAD_ZONE * denoise_percent
            output_attenuation_percent = self.output_attenuator.percent() * 2

            # calculate the previous sample based on our current noise filter size
            prev_volts = self.denoise(n_bins)

            # grab the new sample and add it to the end of the list
            self.bins.pop(0)
            self.bins.append(ain.read_voltage(samples=n_samples))
            curr_volts = self.denoise(n_bins)

            # calculate the change in volts
            d_volts = curr_volts - prev_volts
            slope = d_volts / self.MAX_SLOPE
            if slope > 1:
                slope = 1.0
            elif slope < -1.0:
                slope = -1.0

            # Set the outputs
            if d_volts > deadzone:
                # ain rising
                self.ain_rising_gate.on()
                self.ain_falling_gate.off()
                self.ain_steady_gate.off()

                self.ain_pos_slope_cv.voltage(slope * output_attenuation_percent * europi_config.MAX_OUTPUT_VOLTAGE)
                self.ain_neg_slope_cv.off()
                self.ain_slope_magnitude_cv.voltage(slope * output_attenuation_percent * europi_config.MAX_OUTPUT_VOLTAGE)
            elif d_volts < -deadzone:
                # ain falling
                self.ain_rising_gate.off()
                self.ain_falling_gate.on()
                self.ain_steady_gate.off()

                self.ain_pos_slope_cv.off()
                self.ain_neg_slope_cv.voltage(-slope * output_attenuation_percent * europi_config.MAX_OUTPUT_VOLTAGE)
                self.ain_slope_magnitude_cv.voltage(-slope * output_attenuation_percent * europi_config.MAX_OUTPUT_VOLTAGE)
            else:
                # ain steady
                self.ain_rising_gate.off()
                self.ain_falling_gate.off()
                self.ain_steady_gate.on()

                self.ain_pos_slope_cv.off()
                self.ain_neg_slope_cv.off()
                self.ain_slope_magnitude_cv.off()

if __name__ == "__main__":
    Slopes().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
 A script meant to recreate the Music Thing Modular Turning Machine Random Sequencer as faithfully
 as possible on the EuroPi hardware using bit shift operations to mimic the analog shift register.

din - clock
ain - cv control over the big knob, added to the knobs value
k1 - the big knob (probability that the sequence changes)
k2 - output scale (0-10v)  or sequence length (2-16 steps)
b1 - write (clear bits) (or config value write_value)
b2 - change k2 function
cv1 - pulse 1 (or config value cv1_pulse_bit)
cv2 - pulse 2 (or config value cv2_pulse_bit)
cv3 - pulse 4 (or config value cv3_pulse_bit)
cv4 - pulse cv1 & cv2
cv5 - pulse cv2 & cv3
cv6 - sequence out

If you'd like to use different bits for the pulse outputs you can update the `CVX_PULSE_BIT`
constants below.

The length, scale, and bit pattern are saved whenever the knob 2 state is changed, or when the user
exits to the menu.
"""
from random import getrandbits, randint
from time import sleep


try:
    from firmware import europi
    from firmware.europi import clamp, MAX_UINT16
    from firmware.europi import din, ain, k1, k2, b1, b2, cv1, cv2, cv3, cv4, cv5, cv6, oled
    from firmware.experimental.knobs import KnobBank
except ImportError:
    import europi
    from europi import clamp, MAX_UINT16
    from europi import din, ain, k1, k2, b1, b2, cv1, cv2, cv3, cv4, cv5, cv6, oled
    from experimental.knobs import KnobBank

import configuration
from europi_script import EuroPiScript

INT_MAX_8 = 0xFF
DEFAULT_BIT_COUNT = 16
MAX_OUTPUT_VOLTAGE = europi.MAX_OUTPUT_VOLTAGE


class TuringMachine:
    """A class meant to recreate the Music Thing Modular Turning Machine Random Sequencer as faithfully as possible in
    micropython using bit shift operations to mimic the TM's analog shift register.

    The `TuringMachine` class keeps its state in several internal member variables which are accessed via property
    variables. See `flip_probability`, `scale`, `length` and `write`. users of the class can modify these
    variables in order to configure the TM's behavior. In many cases script authors will want to tie one of these
    variables to a hardware UI element, for example, assigning the `flip_probability` to a knob. In this case clients
    can override the corresponding `*_getter` function. For example, to tie the write switch to button 1::

        tm = TuringMachine()

        def write(self):
          return b1.value()

        tm.write_getter = write

    This form allows clients to override where the `TuringMachine` obtains some of its state variables without
    explicitly sub-classing the `TuringMachine` class. See `EuroPiTuringMachine` for a more detailed example.
    """

    def __init__(
        self,
        bit_count=DEFAULT_BIT_COUNT,
        max_output_voltage=MAX_OUTPUT_VOLTAGE,
        clear_on_write=True,
        flip_probability=0,
        scale=MAX_OUTPUT_VOLTAGE,
        length=DEFAULT_BIT_COUNT,
    ):
        """Create a new TuringMachine with a shift register of the specified bit count. Default is 16, minimum is 8.
        The maximum output voltage is also configurable and defaults to `europi.MAX_OUTPUT_VOLTAGE`
        """

        if bit_count < 8:
            raise ValueError(f"Specified bit_count ({bit_count}) is less than the minimum (8).")
        self.bit_count = bit_count
        self.bits = getrandbits(self.bit_count)
        self._flip_probability = flip_probability
        self.max_output_voltage = max_output_voltage
        self.clear_on_write = clear_on_write
        self._scale = scale
        self._length = length
        self._write = False

        self.flip_probability_getter = lambda: self._flip_probability
        self.scale_getter = lambda: self._scale
        self.length_getter = lambda: self._length
        self.write_getter = lambda: self._write
        self.step_handler = lambda: None

    def _rotate_bits(self):
        self.bits = ((self.bits << 1) % (1 << self.bit_count)) | (
            (self.bits >> (self.length - 1)) & 1
        )

    def step(self):
        """Move the turing machine to its next state based on its current state. Parameters that affect the next state
        include: `flip_probability`, `length`, `write`, and, the internal bit register.

        Typically this method would be called in response to a clock tick.
        """
        self._rotate_bits()
        if self.write:
            if self.clear_on_write:
                self.bits = self.bits & ~1
            else:
                self.bits = self.bits | (1 << 0)

        if randint(0, 99) < self.flip_probability:
            self.bits = self.bits ^ 0b1
        self.step_handler()

    def get_8_bits(self):
        """Returns the least significant eight bits from the internal bit register, which are the same bits used to
        determine the voltage returned by `get_voltage()`. This method is useful when displaying the current state in a
        UI."""
        return self.bits & 0xFF

    def get_bit(self, i):
        """Returns the bit at the specified index. This method exists to support the Pulses Expander functionality."""
        return self.bits >> i & 1

    def get_bit_and(self, *args):
        """Returns the result of a bitwise and of the bits at the specified indexes. This method exists to support the
        Pulses Expander functionality."""
        result = 1
        for i in args:
            result = result & self.get_bit(i)
        return result

    def get_voltage(self):
        """Returns the voltage described by the eight least significant bits of the internal bit register, scaled by the
        current `scale` factor."""
        return self.get_8_bits() / INT_MAX_8 * self.scale

    @property
    def flip_probability(self):
        """Returns the probability that the most significant bit will be flipped when it is rotated into the least
        significant bit's position. This translates to the probability that the sequence will change. The
        flip_probability is represented as an integer in the range [0, 100].
        """
        return self.flip_probability_getter()

    @flip_probability.setter
    def flip_probability(self, probability: int):
        """Set the flip probability as an integer in the range [0, 100]."""
        if probability < 0 or probability > 100:
            raise ValueError(
                f"Probability of {probability} is outside the expected range of [0,100]"
            )
        self._flip_probability = probability

    @property
    def scale(self):
        """Returns the current scaling factor, used to reduce the range of the output voltage to something lower than
        the `max_output_voltage`. Represented by a float in the range [0, `max_output_voltage`]"""
        return self.scale_getter()

    @scale.setter
    def scale(self, scale):
        """Set the scale factor as a float in the range [0, `max_output_voltage`]"""
        if scale < 0 or scale > self.max_output_voltage:
            raise ValueError(
                f"Scale of {scale} is outside the expected range of [0,{self.max_output_voltage}]"
            )
        self._scale = scale

    @property
    def length(self):
        """Returns the length of the current sequence as an integer in the range of [2, `bit_count`]"""
        return self.length_getter()

    @length.setter
    def length(self, length):
        """Sets the length of the current sequence as an integer in the range of [2, `bit_count`]"""
        if length < 2 or length > self.bit_count:
            raise ValueError(
                f"Length of {length} is outside the expected range of [2,{self.bit_count}]"
            )
        self._length = length

    @property
    def write(self):
        """Returns the current value of the 'write switch'. When true the least significant bit will be cleared during
        rotation, regardless of the `flip_probability`. This allows for real-time user manipulation of the sequence.
        """
        return self.write_getter()

    @write.setter
    def write(self, value: bool):
        """Set the state of the 'write switch'. `True` means that the least significant bit will be cleared."""
        self._write = value


class EuroPiTuringMachine(EuroPiScript):
    def __init__(self, bit_count=DEFAULT_BIT_COUNT, max_output_voltage=MAX_OUTPUT_VOLTAGE):
        super().__init__()

        self.LENGTH_CHOICES = [2, 3, 4, 5, 6, 8, 12, 16]  # TODO: vary based on bit_count?
        initial_scale_percent = 0.5  # TODO: load from saved state
        initial_length = 8  # TODO: load from saved state

        self.tm = TuringMachine(
            bit_count=bit_count,
            max_output_voltage=max_output_voltage,
            clear_on_write=self.config.WRITE_VALUE == 0,
            length=initial_length,
            scale=MAX_OUTPUT_VOLTAGE * initial_scale_percent,
        )
        self.tm.flip_probability_getter = self.flip_probability
        self.tm.scale_getter = self.scale
        self.tm.length_getter = self.length
        self.tm.write_getter = self.write
        self.tm.step_handler = self.step_handler
        self.request_next_k2 = False
        self.kb2 = (
            KnobBank.builder(k2)
            .with_disabled_knob()
            .with_locked_knob("scale", initial_percentage_value=initial_scale_percent)
            .with_locked_knob(
                "length",
                initial_percentage_value=(self.LENGTH_CHOICES.index(initial_length) * 2 + 1)
                / (len(self.LENGTH_CHOICES) * 2),
            )
            .build()
        )

        self.cv1_pulse_bit = self.config.CV1_PULSE_BIT
        self.cv2_pulse_bit = self.config.CV2_PULSE_BIT
        self.cv3_pulse_bit = self.config.CV3_PULSE_BIT

        @din.handler
        def clock():
            self.tm.step()

        @b2.handler_falling
        def request_next_k2_mode():
            self.request_next_k2 = True

    def next_k2_mode(self):
        if self.kb2.current_name == "scale":
            self.tm.scale = self.scale()
        elif self.kb2.current_name == "length":
            self.tm.length = self.length()
        self.kb2.next()
        self.request_next_k2 = False

    def step_handler(self):
        cv1.value(self.tm.get_bit(self.cv1_pulse_bit))
        cv2.value(self.tm.get_bit(self.cv2_pulse_bit))
        cv3.value(self.tm.get_bit(self.cv3_pulse_bit))
        cv4.value(self.tm.get_bit_and(self.cv1_pulse_bit, self.cv2_pulse_bit))
        cv5.value(self.tm.get_bit_and(self.cv2_pulse_bit, self.cv3_pulse_bit))
        cv6.voltage(self.tm.get_voltage())

    def flip_probability(self):
        return clamp(int(round(1 - k1.percent() - ain.percent(), 2) * 100), 0, 100)

    def scale(self):
        if self.kb2.current_name == "scale":
            return self.kb2.scale.percent() * self.tm.max_output_voltage
        else:
            return self.tm._scale

    def length(self):
        if self.kb2.current_name == "length":
            return self.kb2.length.choice(self.LENGTH_CHOICES)
        else:
            return self.tm._length

    def write(self):
        return b1.value()

    @staticmethod
    def bits_as_led_line(oled, bits):
        bit_str = f"{bits:08b}"
        x_pos = 0
        width = int(europi.OLED_WIDTH / 8)
        for c in bit_str:
            if c == "1":
                oled.hline(x_pos, 0, width - 1, 1)
            x_pos += width

    @classmethod
    def display_name(cls):
        return "Turing Machine"

    @classmethod
    def config_points(cls):
        range_min = 0
        range_max = min(DEFAULT_BIT_COUNT, 8) - 1

        return [
            configuration.choice(name="WRITE_VALUE", choices=[0, 1], default=0),
            # simulate the actual bits available in the pulses expander (1-7)
            configuration.integer(name="CV1_PULSE_BIT", minimum=range_min, maximum=range_max, default=1),
            configuration.integer(name="CV2_PULSE_BIT", minimum=range_min, maximum=range_max, default=2),
            configuration.integer(name="CV3_PULSE_BIT", minimum=range_min, maximum=range_max, default=4),
        ]

    def main(self):
        line1_y = 11
        line2_y = 23
        while True:
            if self.request_next_k2:
                self.next_k2_mode()

            oled.fill(0)
            prob = self.tm.flip_probability
            prob_2 = (
                "locked"
                if self.tm.flip_probability == 0
                else "looped"
                if self.tm.flip_probability == 100
                else ""
            )
            scale_str = (
                f"{'*' if self.kb2.current_name == 'scale' else ' '}scale:{self.tm.scale:3.1f}"
            )
            len_str = f"{'*' if self.kb2.current_name == 'length' else ' '}{self.tm.length:2} steps"

            self.bits_as_led_line(oled, self.tm.get_8_bits())

            oled.text(f" {prob}%", 0, line1_y, 1)
            oled.text(f"{scale_str}", 40, line1_y, 1)

            oled.text(f"{prob_2}", 0, line2_y, 1)
            oled.text(f"{len_str}", 63, line2_y, 1)
            oled.show()


if __name__ == "__main__":
    EuroPiTuringMachine(DEFAULT_BIT_COUNT, europi.MAX_OUTPUT_VOLTAGE).main()



# Copyright 2025 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A bunch of pixels bouncing around, inspired by the beloved bouncing DVD logo.
A gate is triggered each time a wall is hit. See `bouncing_pixels.md` for more information.

@author  Jorin (jorin.se)
@year    2025
"""

try:
    # Local development
    from software.firmware.europi import *
    from software.firmware.europi_script import EuroPiScript
    from software.firmware.experimental.math_extras import rescale
    from software.firmware.experimental.knobs import KnobBank
    from software.firmware.experimental.thread import DigitalInputHelper
    from software.firmware import configuration

except ImportError:
    # Device import paths
    from europi import *  # type: ignore
    from europi_script import EuroPiScript  # type: ignore
    from experimental.math_extras import rescale  # type: ignore
    from experimental.knobs import KnobBank  # type: ignore
    from experimental.thread import DigitalInputHelper  # type: ignore
    import configuration  # type: ignore

from _thread import start_new_thread, allocate_lock
from cmath import phase, polar, rect
from math import degrees, e, inf, log, pi, radians
from random import uniform
from time import ticks_ms, ticks_diff, sleep_ms

tau = pi * 2

ARENA_HEIGHT = 480
UNITS_PER_PIXEL = ARENA_HEIGHT / oled.height

COLLISION_ID_UP = 0
COLLISION_ID_LEFT = 1
COLLISION_ID_RIGHT = 2
COLLISION_ID_DOWN = 3
COLLISION_ID_ANY = 4
COLLISION_ID_CORNER = 5


def exponential_interpolation(a, b, t):
    """Exponentially interpolate between a and b by t, where t=0 gives a and t=1 gives b.
    This function uses e as the base which hopefully gives a "natural" feeling curve.
    """
    return e ** (t * log(b + 1 - a)) - 1 + a


class Event:
    """A simple, synchronous event implementation. Handlers can be added either when instantiating,
    or by using the += operator. The event is then emitted using the emit function.
    """

    def __init__(self, *handlers):
        # *handlers is a tuple, so we must explicitly create a mutable list to add given handlers to.
        self.handlers = []
        self.handlers.extend(handlers)

    def __iadd__(self, handler):
        """Register an event handling function."""
        self.handlers.append(handler)
        return self

    def emit(self, *args, **kwargs):
        """Emit the event, calling each handler synchronously in the order they were added.
        Arguments are passed through to each handler, so it is up to the user to ensure that
        the arguments match up.
        """
        for handler in self.handlers:
            handler(*args, **kwargs)


class Arena:
    def __init__(self, config):
        self.config = config
        self.on_width_changed = Event()

        # Height is constant, unlike width.
        self.height = ARENA_HEIGHT  # Simulated size
        self.display_height = oled.height  # Displayed arena height
        self.draw_y_min = 0  # First y coordinate for the displayed arena
        self.draw_y_max = (  # Last y coordinate for the displayed arena
            self.draw_y_min + self.display_height
        )

        # Width is variable, but we initialise its variables here to ensure it can be referenced.
        self.width = config.ARENA_WIDTH_MAX
        self.display_width = oled.width
        self.draw_x_min = 0
        self.draw_x_max = oled.width

    def set_width(self, width):
        """Set the arena's width."""
        prev_width = self.width
        self.width = clamp(width, self.config.ARENA_WIDTH_MIN, self.config.ARENA_WIDTH_MAX)
        self.display_width = int(self.width / UNITS_PER_PIXEL)
        self.draw_x_min = int(oled.width / 2 - self.display_width / 2)
        self.draw_x_max = self.draw_x_min + self.display_width
        self.on_width_changed.emit(prev_width, self.width)

    def draw_boundary(self):
        """Draw the boundaries of the arena, one pixel outside of draw_x_min and draw_x_max.
        This will fall outside of the bounds of the display when the arena covers the full screen, which is intentional.
        """
        oled.vline(self.draw_x_min - 1, 0, oled.height, 1)
        oled.vline(self.draw_x_max + 1, 0, oled.height, 1)


class Ball:
    def __init__(self, arena, config):
        self.config = config
        self.BOUNCE_ANGLE_DEVIATION_MAX = radians(config.BOUNCE_ANGLE_DEVIATION_MAX)
        self.arena = arena

        self.on_over_speed = Event(getattr(self, config.OVER_SPEED_BEHAVIOUR))
        self.on_under_speed = Event(getattr(self, config.UNDER_SPEED_BEHAVIOUR))
        self.on_collide = Event()
        arena.on_width_changed += self.translate_x

        self.reset()

    def __repr__(self):
        speed, direction = polar(self.velocity)
        return f"Ball at ({self.pos.real}, {self.pos.imag}), velocity {speed}/s {int(degrees(direction))}°"

    def if_active(fn):
        """Functions wrapped with this function will only run if self.active is true.
        This is needed to avoid processing when over/under speed behaviour is set to deactivate.
        """

        def inner(self, *args, **kwargs):
            if not self.active:
                return
            fn(self, *args, **kwargs)

        return inner

    @if_active
    def draw(self):
        arena = self.arena
        x = int(rescale(self.pos.real, 0, arena.width, arena.draw_x_min, arena.draw_x_max))
        y = int(rescale(self.pos.imag, 0, arena.height, arena.draw_y_min, arena.draw_y_max))
        oled.pixel(x, y, 1)

    def translate_x(self, old_width: float, new_width: float):
        """Translate the ball's x so that it's the same relative distance from the edges with new_width as it was with old_width."""
        new_x = new_width * self.pos.real / old_width
        self.pos = complex(new_x, self.pos.imag)

    def reset(self):
        """Reset the ball, randomising its position, velocity, bounciness, acceleration, and marking it as active."""
        self.pos = complex(uniform(0, self.arena.width), uniform(0, self.arena.height))
        self.velocity = rect(
            uniform(self.config.START_SPEED_MIN, self.config.START_SPEED_MAX), uniform(0, tau)
        )
        self.bounciness = uniform(self.config.BOUNCINESS_MIN, self.config.BOUNCINESS_MAX)
        self.acceleration = uniform(self.config.ACCEL_MIN, self.config.ACCEL_MAX)
        self.active = True

    @if_active
    def impulse(self):
        """Apply an impulse of speed in a random direction to the ball."""
        self.velocity += rect(
            uniform(
                self.config.IMPULSE_SPEED_VARIATION_MIN, self.config.IMPULSE_SPEED_VARIATION_MAX
            )
            * self.impulse_speed,
            uniform(0, tau),
        )

    def deactivate(self):
        """Set the ball to inactive, meaning it won't be processed by functions that are wrapped by @if_active."""
        self.active = False

    def noop(self):
        """Do nothing. This is used as a possible under speed behaviour."""
        pass

    @if_active
    def tick(self, delta: float):
        # Apply acceleration
        delta_v = rect(self.acceleration * delta, phase(self.velocity))
        self.velocity += delta_v

        # Grab the speed and direction of the resulting velocity for later use
        speed, direction = polar(self.velocity)

        # Apply velocity and check whether collisions happened
        delta_x = self.velocity.real * delta
        delta_y = self.velocity.imag * delta

        collide_x = (self.pos.real + delta_x) // self.arena.width
        collide_y = (self.pos.imag + delta_y) // self.arena.height

        self.pos += complex(
            delta_x - delta_x * abs(collide_x) * 2,
            delta_y - delta_y * abs(collide_y) * 2,
        )

        if collide_x != 0:
            if collide_x != 1:  # Left
                self.on_collide.emit(COLLISION_ID_LEFT)
            if collide_x != -1:  # Right
                self.on_collide.emit(COLLISION_ID_RIGHT)
            if (
                self.pos.imag < self.config.CORNER_COLLISION_MARGIN
                or self.arena.height - self.pos.imag < self.config.CORNER_COLLISION_MARGIN
            ):
                self.on_collide.emit(COLLISION_ID_CORNER)

            speed *= self.bounciness
            direction = (
                pi
                - direction
                + uniform(-self.BOUNCE_ANGLE_DEVIATION_MAX, self.BOUNCE_ANGLE_DEVIATION_MAX)
            )

        if collide_y != 0:
            if collide_y != 1:  # Up
                self.on_collide.emit(COLLISION_ID_UP)
            if collide_y != -1:  # Down
                self.on_collide.emit(COLLISION_ID_DOWN)

            speed *= self.bounciness
            direction = -direction + uniform(
                -self.BOUNCE_ANGLE_DEVIATION_MAX, self.BOUNCE_ANGLE_DEVIATION_MAX
            )

        # Update velocity based on changes to speed and direction stemming from collisions
        self.velocity = rect(speed, direction)

        # Travelled more than one width or height in one tick, or speed is above threshold.
        # TODO: we should be able to account for multiple collisions in one tick and not need this check
        if abs(collide_x) > 1 or abs(collide_y) > 1 or speed > self.config.OVER_SPEED_THRESHOLD:
            self.on_over_speed.emit()

        # Speed too low
        if speed < self.config.UNDER_SPEED_THRESHOLD:
            self.on_under_speed.emit()


class Gate:
    def __init__(self, cv, gate_hold_length):
        self.cv = cv
        self.gate_hold_length = gate_hold_length
        self.opened = None

    def on(self):
        self.cv.on()
        self.opened = ticks_ms()

    def off(self):
        self.cv.off()
        self.opened = None

    def tick(self, delta: float):
        if self.opened is None:
            return
        if ticks_diff(ticks_ms(), self.opened) >= self.gate_hold_length:
            self.off()


class BouncingPixels(EuroPiScript):
    def __init__(self):
        super().__init__()
        self.is_running = False

        # This lock is for both the arena and the balls.
        # Since changing arena size changes ball positions, it is a shared lock.
        self.state_lock = allocate_lock()

        saved_state = self.load_state_json()
        self.state_saved = False
        self.k1_bank = (
            KnobBank.builder(k1)
            .with_unlocked_knob("speed")
            .with_locked_knob(
                "ball_count", initial_percentage_value=saved_state.get("ball_count", 0.0)
            )
            .build()
        )

        self.k2_bank = (
            KnobBank.builder(k2)
            .with_unlocked_knob("width")
            .with_locked_knob("impulse_speed", initial_percentage_value=saved_state.get("impulse_speed", 0.5))
            .build()
        )

        # Initialise some internal state variables
        self.b1_pressed = None
        self.b2_pressed = None
        self.ball_count = 1
        self.time_factor = 1

        # Trackers of inputs
        # -1 as a starting value means they always get updated on first poll.
        self.speed_input = -1.0
        self.width_input = -1.0
        self.ball_count_input = -1.0
        self.impulse_speed_input = -1.0
        self.ain_input = -1.0

        # We store analogue input values in these. Although we only change change one as determined by
        # config.AIN_FUNCTION, we do need to reference them all when their respective knobs change.
        self.speed_ain_term = 0.0
        self.width_ain_term = 0.0
        self.ball_count_ain_term = 0.0
        self.impulse_speed_ain_term = 0.0

        # Initialise input events
        self.on_speed_input = Event(self.mark_state_unsaved, self.apply_speed)
        self.on_width_input = Event(self.mark_state_unsaved, self.apply_width)
        self.on_ball_count_input = Event(self.mark_state_unsaved, self.apply_ball_count)
        self.on_impulse_speed_input = Event(self.mark_state_unsaved, self.apply_impulse_speed)
        self.on_ain_input = Event(
            # To avoid excess compute spent on string formatting or dict lookups in the main loop, we keep an assortment of
            # set and apply functions so that we can assign them once and for all here.
            getattr(self, f"update_{self.config.AIN_FUNCTION}_ain_term"),
            getattr(self, f"apply_{self.config.AIN_FUNCTION}"),
        )

        # Create the playing field and gate abstractions
        self.arena = Arena(self.config)
        self.balls = [
            Ball(arena=self.arena, config=self.config) for _ in range(self.config.BALL_COUNT_MAX)
        ]
        hold_lengths = [
            self.config.GATE_HOLD_LENGTH_TOP,
            self.config.GATE_HOLD_LENGTH_LEFT,
            self.config.GATE_HOLD_LENGTH_RIGHT,
            self.config.GATE_HOLD_LENGTH_BOTTOM,
            self.config.GATE_HOLD_LENGTH_ANY,
            self.config.GATE_HOLD_LENGTH_CORNER,
        ]
        self.gates = [Gate(cv, hold_length) for cv, hold_length in zip(cvs, hold_lengths)]

        # Register external event handlers.
        self.digital_input_helper = DigitalInputHelper(
            on_b1_rising=self.b1_rising,
            on_b1_falling=self.b1_falling,
            on_b2_rising=self.b2_rising,
            on_b2_falling=self.b2_falling,
            on_din_rising=getattr(self, self.config.DIN_FUNCTION)
        )

        for ball in self.balls:
            ball.on_collide += self.report_collision

    @classmethod
    def config_points(cls):
        return [
            configuration.floatingPoint(
                "POLL_FREQUENCY", minimum=5.0, maximum=inf, default=30.0, danger=True
            ),
            configuration.floatingPoint(
                "SAVE_PERIOD", minimum=0.0, maximum=inf, default=5000.0, danger=True
            ),
            configuration.floatingPoint(
                "RENDER_FREQUENCY", minimum=1.0, maximum=inf, default=30.0, danger=True
            ),
            configuration.floatingPoint(
                "LONG_PRESS_LENGTH", minimum=0.0, maximum=inf, default=500.0
            ),
            configuration.floatingPoint("TIMESCALE_MIN", minimum=-inf, maximum=inf, default=0.0),
            configuration.floatingPoint("TIMESCALE_MAX", minimum=-inf, maximum=inf, default=100.0),
            configuration.floatingPoint(
                "KNOB_CHANGE_THRESHOLD", minimum=0.0, maximum=0.1, default=0.01
            ),
            configuration.choice("DIN_FUNCTION", choices=["impulse", "reset"], default="impulse"),
            configuration.choice(
                "AIN_FUNCTION",
                choices=["speed", "impulse_speed", "ball_count", "width"],
                default="speed",
            ),
            configuration.floatingPoint(
                "GATE_HOLD_LENGTH_TOP", minimum=1.0, maximum=10_000.0, default=25.0
            ),
            configuration.floatingPoint(
                "GATE_HOLD_LENGTH_LEFT", minimum=1.0, maximum=10_000.0, default=25.0
            ),
            configuration.floatingPoint(
                "GATE_HOLD_LENGTH_RIGHT", minimum=1.0, maximum=10_000.0, default=25.0
            ),
            configuration.floatingPoint(
                "GATE_HOLD_LENGTH_BOTTOM", minimum=1.0, maximum=10_000.0, default=25.0
            ),
            configuration.floatingPoint(
                "GATE_HOLD_LENGTH_ANY", minimum=1.0, maximum=10_000.0, default=10.0
            ),
            configuration.floatingPoint(
                "GATE_HOLD_LENGTH_CORNER", minimum=1.0, maximum=10_000.0, default=100.0
            ),
            configuration.floatingPoint(
                "ARENA_WIDTH_MIN",
                minimum=ARENA_HEIGHT * 2.0 / oled.height,
                maximum=ARENA_HEIGHT * oled.width / oled.height,
                default=ARENA_HEIGHT * 2.0 / oled.height,
            ),
            configuration.floatingPoint(
                "ARENA_WIDTH_MAX",
                minimum=ARENA_HEIGHT * 2.0 / oled.height,
                maximum=ARENA_HEIGHT * oled.width / oled.height,
                default=ARENA_HEIGHT * oled.width / oled.height,
            ),
            configuration.integer("BALL_COUNT_MIN", minimum=1, maximum=100, default=1),
            configuration.integer("BALL_COUNT_MAX", minimum=1, maximum=100, default=100),
            configuration.floatingPoint(
                "CORNER_COLLISION_MARGIN",
                minimum=0.0,
                maximum=ARENA_HEIGHT / 2.0,
                default=ARENA_HEIGHT / oled.height,
            ),
            configuration.floatingPoint("START_SPEED_MIN", minimum=0.0, maximum=inf, default=10.0),
            configuration.floatingPoint("START_SPEED_MAX", minimum=0.0, maximum=inf, default=100.0),
            configuration.floatingPoint("ACCEL_MIN", minimum=-inf, maximum=inf, default=-5.0),
            configuration.floatingPoint("ACCEL_MAX", minimum=-inf, maximum=inf, default=5.0),
            configuration.floatingPoint("BOUNCINESS_MIN", minimum=0.0001, maximum=inf, default=0.8),
            configuration.floatingPoint("BOUNCINESS_MAX", minimum=0.0001, maximum=inf, default=1.2),
            configuration.floatingPoint(
                "BOUNCE_ANGLE_DEVIATION_MAX", minimum=0.0, maximum=180.0, default=15.0
            ),
            configuration.choice(
                "UNDER_SPEED_BEHAVIOUR",
                choices=["impulse", "reset", "deactivate", "noop"],
                default="reset",
            ),
            configuration.choice(
                "OVER_SPEED_BEHAVIOUR", choices=["reset", "deactivate"], default="reset"
            ),
            configuration.floatingPoint(
                "UNDER_SPEED_THRESHOLD", minimum=0.0, maximum=inf, default=5.0
            ),
            configuration.floatingPoint(
                "OVER_SPEED_THRESHOLD", minimum=0.0, maximum=inf, default=1.0e6
            ),
            configuration.floatingPoint("IMPULSE_SPEED_MIN", minimum=0.0, maximum=inf, default=0.0),
            configuration.floatingPoint(
                "IMPULSE_SPEED_MAX", minimum=0.0, maximum=inf, default=1.0e5
            ),
            configuration.floatingPoint(
                "IMPULSE_SPEED_VARIATION_MIN", minimum=0.0, maximum=inf, default=0.5
            ),
            configuration.floatingPoint(
                "IMPULSE_SPEED_VARIATION_MAX", minimum=0.0, maximum=inf, default=2.0
            ),
        ]

    # Button handlers
    def b1_rising(self):
        self.k1_bank.set_current("ball_count")
        self.k2_bank.set_current("impulse_speed")
        self.b1_pressed = ticks_ms()

    def b1_falling(self):
        self.k1_bank.set_current("speed")
        self.k2_bank.set_current("width")

        delta = ticks_diff(ticks_ms(), self.b1_pressed)
        if delta <= self.config.LONG_PRESS_LENGTH:
            self.reset()

        self.b1_pressed = None

    def b2_rising(self):
        self.k1_bank.set_current("ball_count")
        self.k2_bank.set_current("impulse_speed")
        self.b2_pressed = ticks_ms()

    def b2_falling(self):
        self.k1_bank.set_current("speed")
        self.k2_bank.set_current("width")

        delta = ticks_diff(ticks_ms(), self.b2_pressed)
        if delta <= self.config.LONG_PRESS_LENGTH:
            self.impulse()

        self.b2_pressed = None

    def mark_state_unsaved(self):
        """Denote that the state saved onto storage does not match current knob state."""
        self.state_saved = False

    def save_state(self):
        """Save knob states to storage and denote that storage matches current knob state."""
        self.save_state_json(
            {
                "speed": self.speed_input,
                "width": self.width_input,
                "ball_count": self.ball_count_input,
                "impulse_speed": self.impulse_speed_input,
            }
        )
        self.state_saved = True

    # Event handlers to save analogue input into respective variables
    def update_speed_ain_term(self):
        self.speed_ain_term = ain.percent()

    def update_width_ain_term(self):
        self.width_ain_term = ain.percent()

    def update_ball_count_ain_term(self):
        self.ball_count_ain_term = ain.percent()

    def update_impulse_speed_ain_term(self):
        self.impulse_speed_ain_term = ain.percent()

    # Event handlers to apply knob and analogue input values
    def apply_speed(self):
        input_sum = self.speed_input + self.speed_ain_term
        self.time_factor = exponential_interpolation(
            self.config.TIMESCALE_MIN, self.config.TIMESCALE_MAX, input_sum
        )

    def apply_width(self):
        input_sum = self.width_input + self.width_ain_term
        width = rescale(
            input_sum, 0.0, 1.0, self.config.ARENA_WIDTH_MIN, self.config.ARENA_WIDTH_MAX
        )
        self.arena.set_width(width)

    def apply_ball_count(self):
        input_sum = self.ball_count_input + self.ball_count_ain_term
        self.ball_count = int(
            exponential_interpolation(
                self.config.BALL_COUNT_MIN, self.config.BALL_COUNT_MAX, input_sum
            )
        )

    def apply_impulse_speed(self):
        # Impulse strength is calibrated so that an input of 0 gives 0.1 and an input of 1 gives 100.
        input_sum = self.impulse_speed_input + self.impulse_speed_ain_term
        impulse_speed = exponential_interpolation(
            self.config.IMPULSE_SPEED_MIN, self.config.IMPULSE_SPEED_MAX, input_sum
        )
        for ball in self.balls:
            ball.impulse_speed = impulse_speed

    def report_collision(self, collision_id: int):
        """Open the gate corresponding to the given collision ID, as well as the any gate."""
        self.gates[COLLISION_ID_ANY].on()
        self.gates[collision_id].on()

    def reset(self):
        """Reset all balls and gates."""
        for ball in self.balls:
            ball.reset()
        for gate in self.gates:
            gate.off()

    def impulse(self):
        """Apply an impulse to all balls currently in play."""
        for i in range(self.ball_count):
            self.balls[i].impulse()

    def poll(self):
        """Poll for input and emit corresponding events when changes are detected"""
        new_speed_input = self.k1_bank.speed.percent()
        new_width_input = self.k2_bank.width.percent()
        new_ball_count_input = self.k1_bank.ball_count.percent()
        new_impulse_speed_input = self.k2_bank.impulse_speed.percent()
        new_ain_input = ain.percent()

        with self.state_lock:
            self.digital_input_helper.update()
            # The difference between the new and registered input must exceed the threshold in order to trigger a change.
            # This reduces jitter, but decreases accuracy.
            if abs(new_speed_input - self.speed_input) > self.config.KNOB_CHANGE_THRESHOLD:
                self.speed_input = new_speed_input
                self.on_speed_input.emit()

            if (
                abs(new_impulse_speed_input - self.impulse_speed_input)
                > self.config.KNOB_CHANGE_THRESHOLD
            ):
                self.impulse_speed_input = new_impulse_speed_input
                self.on_impulse_speed_input.emit()

            # Since the rest of the parameters require a button to be held, they will not jitter once the button is released
            # and no threshold check should be needed.
            if new_ball_count_input != self.ball_count_input:
                self.ball_count_input = new_ball_count_input
                self.on_ball_count_input.emit()

            if new_width_input != self.width_input:
                self.width_input = new_width_input
                self.on_width_input.emit()

            if new_ain_input != self.ain_input:
                self.ain_input = new_ain_input
                self.on_ain_input.emit()

    def tick(self, delta: float):
        """Process a tick in the simulation.
        Delta should already include any time scaling, meaning that this function is
        naive to any time scale.
        """
        any_active = False
        with self.state_lock:
            for i in range(self.ball_count):
                self.balls[i].tick(delta)
                any_active = any_active or self.balls[i].active
            if not any_active:
                self.reset()
        for gate in self.gates:
            gate.tick(delta)

    def render(self):
        """Render the display.
        This is the only function that should call any drawing commands.
        """
        oled.fill(0)
        with self.state_lock:
            self.arena.draw_boundary()
            for i in range(self.ball_count):
                if not self.balls[i].active:
                    continue
                self.balls[i].draw()
        oled.show()

    def main(self):
        """Start the application."""
        # Per chrisib's recommendation after some display-related errors,
        # the render thread specifically runs on cpu0 as this might be the
        # cpu that handles i2c communication with the display.
        self.is_running = True
        try:
            proc_thread = start_new_thread(self.proc_thread, ())
            self.render_thread()
        except KeyboardInterrupt:
            self.is_running = False
        finally:
            print("User aborted. Exiting.")

    def proc_thread(self):
        """Run the simulation, poll inputs, and save state as necessary."""
        prev_cycle = None
        last_poll = None
        poll_period = 1000.0 / self.config.POLL_FREQUENCY
        usb_connected_at_start = usb_connected.value()
        while usb_connected.value() == usb_connected_at_start and self.is_running:
            cycle_start = ticks_ms()
            delta = ticks_diff(cycle_start, prev_cycle) / 1000
            
            # Poll inputs at limited frequency
            time_since_poll = ticks_diff(cycle_start, last_poll)
            if time_since_poll > poll_period:
                self.poll()
                last_poll = cycle_start

            # Save at limited frequency
            if not self.state_saved and self.last_saved() > self.config.SAVE_PERIOD:
                self.save_state()

            self.tick(delta * self.time_factor)
            prev_cycle = cycle_start

    def render_thread(self):
        """Render at limited frequency."""
        render_period = 1000.0 / self.config.RENDER_FREQUENCY
        usb_connected_at_start = usb_connected.value()
        while usb_connected.value() == usb_connected_at_start and self.is_running:
            cycle_start = ticks_ms()
            self.render()
            cycle_finish = ticks_ms()
            time_taken = ticks_diff(cycle_finish, cycle_start)
            wait = int(max(0.0, render_period - time_taken))
            sleep_ms(wait)


if __name__ == "__main__":
    BouncingPixels().main()



#!/usr/bin/env python3
# Copyright 2023 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convert incoming triggers to gates or gates to triggers

Also outputs a toggle input which changes state every time an incoming gate is received, or when either button is
pressed.

@author Chris Iverach-Brereton <ve4cib@gmail.com>
@year   2023-4
"""

import time

from europi import *
from europi_script import EuroPiScript
from math import sqrt

from experimental.knobs import MedianAnalogInput
from experimental.screensaver import Screensaver

## Trigger outputs are 10ms long (rising/falling edges of gate signals)
TRIGGER_DURATION_MS = 10

# Gate outputs have a range of 50ms to 1s normally
# Maximum actual value can be increased via CV
MIN_GATE_DURATION_MS = 50
MAX_GATE_DURATION_MS = 1000


class GatesAndTriggers(EuroPiScript):
    def __init__(self):
        super().__init__()

        self.on_incoming_rise_start_time = 0
        self.on_incoming_fall_start_time = 0

        self.ain = MedianAnalogInput(ain)
        self.k1 = MedianAnalogInput(k1)
        self.k2 = MedianAnalogInput(k2)

        # assign each of the CV outputs to specific duties
        self.gate_out = cv1
        self.incoming_rise_out = cv2
        self.incoming_fall_out = cv3
        self.gate_fall_out = cv4
        self.toggle_out = cv5
        self.toggle_fall_out = cv6

        turn_off_all_cvs()

        self.last_rise_at = 0
        self.last_fall_at = 0
        self.force_toggle = False

        self.last_user_interaction_at = 0

        self.screensaver = Screensaver()

        @din.handler
        def on_din_rising():
            self.last_rise_at = time.ticks_ms()

        @din.handler_falling
        def on_din_falling():
            self.last_fall_at = time.ticks_ms()

        @b1.handler
        def on_b1_press():
            now = time.ticks_ms()
            self.last_rise_at = now
            self.last_user_interaction_at = now

        @b1.handler_falling
        def on_b1_release():
            self.last_fall_at = time.ticks_ms()

        @b2.handler
        def on_b2_press():
            self.force_toggle = True
            self.last_user_interaction_at = time.ticks_ms()

    def quadratic_knob(self, x):
        """Some magic math to give us a quadratic response on the knob percentage

        This gives us 50ms @ 0% and 1000ms @ 100% with greater precision at the higher end
        where the differences will be more noticeable

        @param x  The value of the knob in the range [0, 100]
        """
        if x <= 0:
            return MIN_GATE_DURATION_MS

        # /10 == sqrt(x) at maximum value
        return (MAX_GATE_DURATION_MS - MIN_GATE_DURATION_MS)/10 * sqrt(x) + MIN_GATE_DURATION_MS

    def main(self):
        ui_dirty = True
        gate_duration = 0
        gate_on = False
        gate_fall_at = 0
        toggle = False
        toggle_on = False
        toggle_fall_at = 0

        last_k1_percent = 0
        last_k2_percent = 0
        last_gate_duration = 0

        while(True):
            # read the knobs with higher samples
            # keep 1 decimal place
            k1_percent = round(self.k1.percent() * 100)         # 0-100
            k2_percent = round(self.k2.percent() * 100) / 100   # 0-1
            cv_percent = round(self.ain.percent() * 100) / 100  # 0-1

            gate_duration = max(
                round(self.quadratic_knob(k1_percent) + cv_percent * k2_percent * 2000),
                MIN_GATE_DURATION_MS
            )

            # Refresh the GUI if the knobs have moved or the gate duration has changed
            ui_dirty = last_k1_percent != k1_percent or last_k2_percent != k2_percent or last_gate_duration != gate_duration

            now = time.ticks_ms()
            time_since_din_rise = time.ticks_diff(now, self.last_rise_at)
            time_since_din_fall = time.ticks_diff(now, self.last_fall_at)

            # CV1: gate output based on rising edge of din/b1
            if time_since_din_rise >= 0 and time_since_din_rise <= gate_duration:
                self.gate_out.on()
                if not gate_on:
                    gate_on = True
                    toggle = not toggle
            elif gate_on:
                gate_on = False
                gate_fall_at = now
                self.gate_out.off()
            else:
                self.gate_out.off()

            # CV2: trigger output for the rising edge of din/b1
            if time_since_din_rise >= 0 and time_since_din_rise <= TRIGGER_DURATION_MS:
                self.incoming_rise_out.on()
            else:
                self.incoming_rise_out.off()

            # CV3: trigger output for falling edge if din/b1
            if time_since_din_fall >= 0 and time_since_din_fall <= TRIGGER_DURATION_MS:
                self.incoming_fall_out.on()
            else:
                self.incoming_fall_out.off()

            # CV4: trigger output for falling edge of cv1
            time_since_gate_fall = time.ticks_diff(now, gate_fall_at)
            if time_since_gate_fall >= 0 and time_since_gate_fall <= TRIGGER_DURATION_MS:
                self.gate_fall_out.on()
            else:
                self.gate_fall_out.off()

            # CV5: toggle output; flips state every time we get a rising edge on din/b1/b2
            if self.force_toggle:
                toggle = not toggle
                self.force_toggle = False
            if toggle:
                self.toggle_out.on()
                toggle_on = True
            elif not toggle and toggle_on:
                self.toggle_out.off()
                toggle_on = False
                toggle_fall_at = now
            else:
                self.toggle_out.off()

            # CV6: trigger output for falling edge of cv5
            time_since_toggle_fall = time.ticks_diff(now, toggle_fall_at)
            if time_since_toggle_fall >= 0 and time_since_toggle_fall <= TRIGGER_DURATION_MS:
                self.toggle_fall_out.on()
            else:
                self.toggle_fall_out.off()

            last_k1_percent = k1_percent
            last_k2_percent = k2_percent
            last_gate_duration = gate_duration

            if ui_dirty:
                self.last_user_interaction_at = now
                oled.centre_text(f"Gate: {gate_duration}ms")
                ui_dirty = False
            elif time.ticks_diff(now, self.last_user_interaction_at) > self.screensaver.ACTIVATE_TIMEOUT_MS:
                self.screensaver.draw()
                self.last_user_interaction_at = time.ticks_add(now, -self.screensaver.ACTIVATE_TIMEOUT_MS)


if __name__=="__main__":
    GatesAndTriggers().main()



# Copyright 2022 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Smooth Random Voltages
author: Adam Wonak (github.com/awonak)
date: 2022-06-01
labels: random, s&h


Random cv with adjustable slew rate.
Inspired by: https://youtu.be/tupkx3q7Dyw

3 random or analog input s&h voltages with changable slew smoothness.

New voltages assigned upon each digital input trigger. Top row outputs move
towards target voltage according to slew rate set by knob 1. Bottom row outputs
immediately change to new target voltage. Voltage source toggled by button 2,
which will switch between random and analog input voltage source. Button 1 will
toggle various display types for the oled.

Col 1, change on each trigger.
Col 2, change on each trigger if target voltage has been reached.
Col 3, change on trigger with knob 2 probability.

digital_input: trigger to set a new target voltage.
analog_input: source for new target value when enabled.

knob_1: slew rate
knob_2: probability of new Col 3 voltage

button_1: cycle through the visualizations: Bars, Scope, Blank.
button_2: cycle through voltage sources: random, analog input.

output_1: smooth random voltage 1 - new voltage on trigger
output_2: smooth random voltage 2 - new voltage after target reached
output_3: smooth random voltage 3 - new voltage based on k2 probability

output_4: stepped random voltage 1 - new voltage on trigger (no slew)
output_5: stepped random voltage 2 - new voltage after target reached (no slew)
output_6: stepped random voltage 3 - new voltage based on k2 probability (no slew)

"""
from random import random, uniform
from utime import ticks_diff, ticks_ms
import machine

try:
    from software.firmware import europi
    from software.firmware.europi import CHAR_HEIGHT, OLED_HEIGHT, OLED_WIDTH
    from software.firmware.europi_script import EuroPiScript

except ImportError:
    import europi
    from europi import CHAR_HEIGHT, OLED_HEIGHT, OLED_WIDTH
    from europi_script import EuroPiScript


# Script Constants
MENU_DURATION = 1800
MAX_SLEW_RATE = 100_000


def envelope_generator(start=0, target=10, slew_rate=512):
    """
    Creates a generator function returning the absolute value differece between
    the start and target voltages over the given duration in ms.
    """
    cv_diff = abs(target - start)
    cur = min(start, target)
    _target = max(start, target)
    slew_rate = max(slew_rate, 1)  # avoid div by zero
    env_start = ticks_ms()
    while cur < _target:
        time_since_start = ticks_diff(ticks_ms(), env_start)
        env_progress = time_since_start / slew_rate
        cur = env_progress * cv_diff
        cur = europi.clamp(cur, europi.MIN_OUTPUT_VOLTAGE, europi.MAX_OUTPUT_VOLTAGE)
        yield cur


class SmoothRandomVoltages(EuroPiScript):
    def __init__(self):
        self.voltages = [0, 0, 0]
        self.target_voltages = [0, 0, 0]

        # Maximim ms duration for slew rate.
        self.slew_rate = lambda: europi.k1.range(MAX_SLEW_RATE)

        # Visualization display choice.
        self.visualization = 0  # 0: Bars, 1: Scope, 2: Blank.

        # Voltage source choice.
        self.voltage_source = 0  # 0: random, 1: analog input.
        self.voltage_source_display = ["random", "analog"]

        # Envelopes
        self.env = [
            envelope_generator(0, 10),
            envelope_generator(0, 10),
            envelope_generator(0, 10),
        ]

        # Register digital input handler
        @europi.din.handler
        def new_target_voltages():
            """Check if next voltage conditions and update if ready."""
            self.set_target_voltages()

        # Cycle through the visualizations
        @europi.b1.handler
        def change_visualization():
            europi.oled.fill(0)
            europi.oled.show()
            # 0: Bars, 1: Scope, 2: Blank.
            self.visualization = (self.visualization + 1) % 3

        # Toggle between random and analog input voltage source
        @europi.b2.handler
        def change_voltage_source():
            # 0: random, 1: analog input.
            self.voltage_source = (self.voltage_source + 1) % 2

    def set_target_voltages(self):
        """Get next random voltage value."""
        # Col 1, change on each trigger.
        self.target_voltages[0] = self.get_new_voltage()
        self.env[0] = envelope_generator(self.voltages[0], self.target_voltages[0], self.slew_rate())

        # Col 2, change on each trigger if target voltage has been reached.
        if self.voltages[1] == self.target_voltages[1]:
            self.target_voltages[1] = self.get_new_voltage()
            self.env[1] = envelope_generator(
                self.voltages[1], self.target_voltages[1], self.slew_rate()
            )

        # Col 3, change on trigger with knob 2 probability.
        if europi.k2.percent() > random():
            self.target_voltages[2] = self.get_new_voltage()
            self.env[2] = envelope_generator(
                self.voltages[2], self.target_voltages[2], self.slew_rate()
            )

    def get_new_voltage(self):
        """Return a new voltage from analog in or random value."""
        # 0: random, 1: analog input.
        if self.voltage_source == 0:
            return uniform(europi.MIN_OUTPUT_VOLTAGE, europi.MAX_OUTPUT_VOLTAGE)
        else:
            return europi.ain.read_voltage()

    def update_display(self):
        """Show current voltage visualizations."""
        if self.visualization == 0:
            self.display_bars()
        elif self.visualization == 1:
            self.display_scope()
        self.show_menu_header()
        europi.oled.show()

    def show_menu_header(self):
        """When button2 has been pressed, show the current changed source value."""
        if ticks_diff(ticks_ms(), europi.b2.last_pressed()) < MENU_DURATION:
            europi.oled.fill_rect(0, 0, OLED_WIDTH, CHAR_HEIGHT, 1)
            msg = f"Source: {self.voltage_source_display[self.voltage_source]}"
            europi.oled.text(msg, 0, 0, 0)

    def display_bars(self):
        """Draw a bar representing the slew / target for each of the 3 voltages."""
        europi.oled.fill(0)
        for i in range(3):
            x1 = 0
            y1 = int(i * (OLED_HEIGHT / 3)) + 2
            y2 = int(OLED_HEIGHT / 3) - 2
            x2_slew = int((self.voltages[i] / europi.MAX_OUTPUT_VOLTAGE) * OLED_WIDTH)
            x2_target = int((self.target_voltages[i] / europi.MAX_OUTPUT_VOLTAGE) * OLED_WIDTH)
            # Smooth voltage rising
            if self.voltages[i] < self.target_voltages[i]:
                europi.oled.fill_rect(x1, y1, x2_slew, y2, 1)
                europi.oled.rect(x1, y1, x2_target, y2, 1)
            # Smooth voltage falling
            else:
                europi.oled.rect(x1, y1, x2_slew, y2, 1)
                europi.oled.fill_rect(x1, y1, x2_target, y2, 1)

    def display_scope(self):
        """Draw a real-time line representing the slew value for each of the 3 voltages."""
        pixel_x = 0
        pixel_y = europi.OLED_HEIGHT - 1
        europi.oled.scroll(1, 0)
        europi.oled.vline(pixel_x, 0, europi.OLED_HEIGHT, 0)
        europi.oled.pixel(pixel_x, pixel_y - int(self.voltages[0] * (pixel_y / 10)), 1)
        europi.oled.pixel(pixel_x, pixel_y - int(self.voltages[1] * (pixel_y / 10)), 1)
        europi.oled.pixel(pixel_x, pixel_y - int(self.voltages[2] * (pixel_y / 10)), 1)

    def main(self):
        # Start the main loop.
        while True:
            for i in range(len(self.voltages)):
                # Smooth voltage rising
                if self.voltages[i] < self.target_voltages[i]:
                    self.voltages[i] += next(self.env[i])
                    self.voltages[i] = min(self.voltages[i], self.target_voltages[i])

                # Smooth voltage falling
                elif self.voltages[i] > self.target_voltages[i]:
                    self.voltages[i] -= next(self.env[i])
                    self.voltages[i] = max(self.voltages[i], self.target_voltages[i])

                # Set the current smooth / stepped voltage.
                europi.cvs[i].voltage(self.voltages[i])
                europi.cvs[i + 3].voltage(self.target_voltages[i])

            self.update_display()


if __name__ == "__main__":
    srv = SmoothRandomVoltages()
    srv.main()
